import tensorflow as tf

__all__ = ["tf_kron_mult", "tf_masked_triangular_solve"]


def tf_solve_lower_triangular_kron(L, y):
    """ Tensorflow function to solve L x = y
    where L = kron(L[0], L[1] .. L[n-1])
    and L[i] are the lower triangular matrices

    Arguments
    ---------
    L : list of 2-D tensors
        Each element of the list must be a tensorflow tensor and
        must be a lower triangular matrix of dimension n_i x n_i

    y : 1-D or 2-D tensor
        Dimension (n_0*n_1*..n_(m-1)) x p

    Returns
    -------
    x : 1-D or 2-D tensor
        Dimension (n_0*n_1*..n_(m-1)) x p

    """
    n = len(L)
    if n == 1:
        return tf.linalg.triangular_solve(L[0], y)
    else:
        x = y
        na = L[0].get_shape().as_list()[0]
        n_list = tf.stack(
            [tf.cast(tf.shape(input=mat)[0], dtype=tf.float64) for mat in L]
        )
        n_prod = tf.cast(tf.reduce_prod(input_tensor=n_list), dtype=tf.int32)
        nb = tf.cast(n_prod / na, dtype=tf.int32)
        col = tf.shape(input=x)[1]

        for i in range(na):
            xt, xinb, xina = tf.split(x, [i * nb, nb, (na - i - 1) * nb], 0)
            t = xinb / L[0][i, i]
            xinb = tf_solve_lower_triangular_kron(L[1:], t)
            xina = xina - tf.reshape(
                tf.tile(tf.slice(L[0], [i + 1, i],
                                 [na - i - 1, 1]), [1, nb * col]),
                [(na - i - 1) * nb, col],
            ) * tf.reshape(
                tf.tile(tf.reshape(t, [-1, 1]), [na - i - 1, 1]),
                [(na - i - 1) * nb, col],
            )
            x = tf.concat(axis=0, values=[xt, xinb, xina])

        return x


def tf_solve_upper_triangular_kron(L, y):
    """ Tensorflow function to solve L^T x = y
    where L = kron(L[0], L[1] .. L[n-1])
    and L[i] are the lower triangular matrices

    Arguments
    ---------
    L : list of 2-D tensors
        Each element of the list must be a tensorflow tensor and
        must be a lower triangular matrix of dimension n_i x n_i

    y : 1-D or 2-D tensor
        Dimension (n_0*n_1*..n_(m-1)) x p

    Returns
    -------
    x : 1-D or 2-D tensor
        Dimension (n_0*n_1*..n_(m-1)) x p

    """
    n = len(L)
    if n == 1:
        return tf.linalg.triangular_solve(L[0], y, adjoint=True)
    else:
        x = y
        na = L[0].get_shape().as_list()[0]
        n_list = tf.stack(
            [tf.cast(tf.shape(input=mat)[0], dtype=tf.float64) for mat in L]
        )
        n_prod = tf.cast(tf.reduce_prod(input_tensor=n_list), dtype=tf.int32)
        nb = tf.cast(n_prod / na, dtype=tf.int32)
        col = tf.shape(input=x)[1]

        for i in range(na - 1, -1, -1):
            xt, xinb, xina = tf.split(x, [i * nb, nb, (na - i - 1) * nb], 0)
            t = xinb / L[0][i, i]
            xinb = tf_solve_upper_triangular_kron(L[1:], t)
            xt = xt - tf.reshape(
                tf.tile(tf.transpose(a=tf.slice(
                    L[0], [i, 0], [1, i])), [1, nb * col]),
                [i * nb, col],
            ) * tf.reshape(tf.tile(tf.reshape(t, [-1, 1]), [i, 1]),
                           [i * nb, col])
            x = tf.concat(axis=0, values=[xt, xinb, xina])

        return x


def tf_kron_mult(L, x):
    """ Tensorflow multiply with kronecker product matrix
    Returns kron(L[0], L[1] ...) * x

    Arguments
    ---------
    L : list of 2-D tensors
        Each element of the list must be a tensorflow tensor and
        must be a square matrix of dimension n_i x n_i

    x : 1-D or 2-D tensor
        Dimension (n_0*n_1*..n_(m-1)) x p

    Returns
    -------
    y : 1-D or 2-D tensor
        Dimension (n_0*n_1*..n_(m-1)) x p
    """
    n = len(L)
    if n == 1:
        return tf.matmul(L[0], x)
    else:
        na = L[0].get_shape().as_list()[0]
        n_list = tf.stack(
            [tf.cast(tf.shape(input=mat)[0], dtype=tf.float64) for mat in L]
        )
        n_prod = tf.cast(tf.reduce_prod(input_tensor=n_list), dtype=tf.int32)
        nb = tf.cast(n_prod / na, dtype=tf.int32)
        col = tf.shape(input=x)[1]
        xt = tf_kron_mult(
            L[1:], tf.transpose(a=tf.reshape(tf.transpose(a=x), [-1, nb]))
        )
        y = tf.zeros_like(x)
        for i in range(na):
            ya, yb, yc = tf.split(y, [i * nb, nb, (na - i - 1) * nb], 0)
            yb = tf.reshape(
                tf.matmul(
                    tf.reshape(xt, [nb * col, na]),
                    tf.transpose(a=tf.slice(L[0], [i, 0], [1, na])),
                ),
                [nb, col],
            )
            y = tf.concat(axis=0, values=[ya, yb, yc])
        return y


def tf_masked_triangular_solve(L, y, mask, lower=True, adjoint=False):
    """ Tensorflow function to solve L x = y
    where L is a lower triangular matrix with a mask

    Arguments
    ---------
    L : 2-D tensor
        Must be a tensorflow tensor and
        must be a triangular matrix of dimension n x n

    y : 1-D or 2-D tensor
        Dimension n x p

    mask : 1-D tensor
        Dimension n x 1, should be 1 if element is valid, 0 if invalid

    lower : boolean (default : True)
        True if L is lower triangular, False if upper triangular

    adjoint : boolean (default : False)
        True if solving for L^T x = y, False if solving for Lx = y

    Returns
    -------
    x : 1-D or 2-D tensor
        Dimension n x p, values at rows for which mask == 0 are set to zero

    """

    zero = tf.constant(0, dtype=tf.int32)
    mask_mat = tf.compat.v1.where(
        tf.not_equal(
            tf.matmul(tf.reshape(mask, [-1, 1]),
                      tf.reshape(mask, [1, -1])), zero
        )
    )
    q = tf.cast(
        tf.sqrt(tf.cast(tf.shape(input=mask_mat)[0], dtype=tf.float64)),
        dtype=tf.int32
    )
    L_masked = tf.reshape(tf.gather_nd(L, mask_mat), [q, q])

    maskindex = tf.compat.v1.where(tf.not_equal(mask, zero))
    y_masked = tf.gather_nd(y, maskindex)

    x_s1 = tf.linalg.triangular_solve(
        L_masked, y_masked, lower=lower, adjoint=adjoint)
    x = tf.scatter_nd(maskindex, x_s1, tf.cast(
        tf.shape(input=y), dtype=tf.int64))
    return x


def tf_solve_lower_triangular_masked_kron(L, y, mask):
    """ Tensorflow function to solve L x = y
    where L = kron(L[0], L[1] .. L[n-1]),
    L[i] are the lower triangular matrices,
    and mask is a binary elementwise mask on the full L

    Arguments
    ---------
    L : list of 2-D tensors
        Each element of the list must be a tensorflow tensor and
        must be a lower triangular matrix of dimension n_i x n_i

    y : 1-D or 2-D tensor
        Dimension [n_0*n_1*..n_(m-1)), p]

    mask: 1-D tensor
        Dimension [n_0*n_1*...n_(m-1)] with 1 for valid rows and 0
        for don't care

    Returns
    -------
    x : 1-D or 2-D tensor
        Dimension (n_0*n_1*..n_(m-1)) x p, values at rows
        for which mask == 0 are set to zero

    """
    n = len(L)
    if n == 1:
        return tf_masked_triangular_solve(L[0], y, mask, lower=True,
                                          adjoint=False)
    else:
        x = y
        na = L[0].get_shape().as_list()[0]
        n_list = tf.stack(
            [tf.cast(tf.shape(input=mat)[0], dtype=tf.float64) for mat in L]
        )
        n_prod = tf.cast(tf.reduce_prod(input_tensor=n_list), dtype=tf.int32)
        nb = tf.cast(n_prod / na, dtype=tf.int32)
        col = tf.shape(input=x)[1]

        for i in range(na):
            mask_b = tf.slice(mask, [i * nb], [nb])
            xt, xinb, xina = tf.split(x, [i * nb, nb, (na - i - 1) * nb], 0)
            t = xinb / L[0][i, i]

            if tf.reduce_sum(input_tensor=mask_b) != nb:
                xinb = tf_solve_lower_triangular_masked_kron(L[1:], t, mask_b)
                t_masked = tf_kron_mult(L[1:], xinb)

            else:
                # all valid - same as no mask
                xinb = tf_solve_lower_triangular_kron(L[1:], t)
                t_masked = t
            xina = xina - tf.reshape(
                tf.tile(tf.slice(L[0], [i + 1, i],
                                 [na - i - 1, 1]), [1, nb * col]),
                [(na - i - 1) * nb, col],
            ) * tf.reshape(
                tf.tile(tf.reshape(t_masked, [-1, 1]), [na - i - 1, 1]),
                [(na - i - 1) * nb, col],
            )

            x = tf.concat(axis=0, values=[xt, xinb, xina])

        return x


def tf_solve_upper_triangular_masked_kron(L, y, mask):
    """ Tensorflow function to solve L^T x = y
    where L = kron(L[0], L[1] .. L[n-1])
    and L[i] are the lower triangular matrices

    Arguments
    ---------
    L : list of 2-D tensors
        Each element of the list must be a tensorflow tensor and
        must be a lower triangular matrix of dimension n_i x n_i

    y : 1-D or 2-D tensor
        Dimension [n_0*n_1*..n_(m-1)), p]

    mask: 1-D tensor
        Dimension [n_0*n_1*...n_(m-1)] with 1 for valid rows
        and 0 for don't care

    Returns
    -------
    x : 1-D or 2-D tensor
        Dimension (n_0*n_1*..n_(m-1)) x p, values at rows
        for which mask == 0 are set to zero

    """
    n = len(L)
    if n == 1:
        return tf_masked_triangular_solve(L[0], y, mask, lower=True,
                                          adjoint=True)
    else:
        x = y
        na = L[0].get_shape().as_list()[0]
        n_list = tf.stack(
            [tf.cast(tf.shape(input=mat)[0], dtype=tf.float64) for mat in L]
        )
        n_prod = tf.cast(tf.reduce_prod(input_tensor=n_list), dtype=tf.int32)
        nb = tf.cast(n_prod / na, dtype=tf.int32)
        col = tf.shape(input=x)[1]
        L1_end_tr = [tf.transpose(a=x) for x in L[1:]]

        for i in range(na - 1, -1, -1):
            mask_b = tf.slice(mask, [i * nb], [nb])
            xt, xinb, xina = tf.split(x, [i * nb, nb, (na - i - 1) * nb], 0)
            t = xinb / L[0][i, i]

            if tf.reduce_sum(input_tensor=mask_b) != nb:
                xinb = tf_solve_upper_triangular_masked_kron(L[1:], t, mask_b)
                t_masked = tf_kron_mult(L1_end_tr, xinb)
            else:
                xinb = tf_solve_upper_triangular_kron(L[1:], t)
                t_masked = t

            xt = xt - tf.reshape(
                tf.tile(tf.transpose(a=tf.slice(
                    L[0], [i, 0], [1, i])), [1, nb * col]),
                [i * nb, col],
            ) * tf.reshape(
                tf.tile(tf.reshape(t_masked, [-1, 1]), [i, 1]), [i * nb, col]
            )
            x = tf.concat(axis=0, values=[xt, xinb, xina])

        return x



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Full Correlation Matrix Analysis (FCMA)

Correlation related high performance routines
"""

# Authors: Yida Wang
# (Intel Labs), 2017

import numpy as np
from . import cython_blas as blas  # type: ignore
from scipy.stats.mstats import zscore
import math

__all__ = [
    "compute_correlation",
]


def _normalize_for_correlation(data, axis, return_nans=False):
    """normalize the data before computing correlation

    The data will be z-scored and divided by sqrt(n)
    along the assigned axis

    Parameters
    ----------
    data: 2D array

    axis: int
        specify which dimension of the data should be normalized

    return_nans: bool, default:False
        If False, return zeros for NaNs; if True, return NaNs

    Returns
    -------
    data: 2D array
        the normalized data
    """
    shape = data.shape
    data = zscore(data, axis=axis, ddof=0)
    # if zscore fails (standard deviation is zero),
    # optionally set all values to be zero
    if not return_nans:
        data = np.nan_to_num(data)
    data = data / math.sqrt(shape[axis])
    return data


def compute_correlation(matrix1, matrix2, return_nans=False):
    """compute correlation between two sets of variables

    Correlate the rows of matrix1 with the rows of matrix2.
    If matrix1 == matrix2, it is auto-correlation computation
    resulting in a symmetric correlation matrix.
    The number of columns MUST agree between set1 and set2.
    The correlation being computed here is
    the Pearson's correlation coefficient, which can be expressed as

    .. math:: corr(X, Y) = \\frac{cov(X, Y)}{\\sigma_X\\sigma_Y}

    where cov(X, Y) is the covariance of variable X and Y, and

    .. math:: \\sigma_X

    is the standard deviation of variable X

    Reducing the correlation computation to matrix multiplication
    and using BLAS GEMM API wrapped by Scipy can speedup the numpy built-in
    correlation computation (numpy.corrcoef) by one order of magnitude

    .. math::
        corr(X, Y)
        &= \\frac{\\sum\\limits_{i=1}^n (x_i-\\bar{x})(y_i-\\bar{y})}{(n-1)
        \\sqrt{\\frac{\\sum\\limits_{j=1}^n x_j^2-n\\bar{x}}{n-1}}
        \\sqrt{\\frac{\\sum\\limits_{j=1}^{n} y_j^2-n\\bar{y}}{n-1}}}\\\\
        &= \\sum\\limits_{i=1}^n(\\frac{(x_i-\\bar{x})}
        {\\sqrt{\\sum\\limits_{j=1}^n x_j^2-n\\bar{x}}}
        \\frac{(y_i-\\bar{y})}{\\sqrt{\\sum\\limits_{j=1}^n y_j^2-n\\bar{y}}})

    By default (return_nans=False), returns zeros for vectors with NaNs.
    If return_nans=True, convert zeros to NaNs (np.nan) in output.

    Parameters
    ----------
    matrix1: 2D array in shape [r1, c]
        MUST be continuous and row-major

    matrix2: 2D array in shape [r2, c]
        MUST be continuous and row-major

    return_nans: bool, default:False
        If False, return zeros for NaNs; if True, return NaNs

    Returns
    -------
    corr_data: 2D array in shape [r1, r2]
        continuous and row-major in np.float32
    """
    matrix1 = matrix1.astype(np.float32)
    matrix2 = matrix2.astype(np.float32)
    [r1, d1] = matrix1.shape
    [r2, d2] = matrix2.shape
    if d1 != d2:
        raise ValueError('Dimension discrepancy')
    # preprocess two components
    matrix1 = _normalize_for_correlation(matrix1, 1,
                                         return_nans=return_nans)
    matrix2 = _normalize_for_correlation(matrix2, 1,
                                         return_nans=return_nans)
    corr_data = np.empty((r1, r2), dtype=np.float32, order='C')
    # blas routine is column-major
    blas.compute_single_matrix_multiplication('T', 'N',
                                              r2, r1, d1,
                                              1.0,
                                              matrix2, d2,
                                              matrix1, d1,
                                              0.0,
                                              corr_data,
                                              r2)
    return corr_data



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import pytest

import brainiak.funcalign.sssrm


def test_instance():
    model = brainiak.funcalign.sssrm.SSSRM()
    assert model, "Invalid SSSRM instance!"


def test_wrong_input():
    from sklearn.utils.validation import NotFittedError
    import numpy as np

    voxels = 100
    align_samples = 400
    samples = 500
    subjects = 2
    features = 3
    n_labels = 4

    model = brainiak.funcalign.sssrm.SSSRM(n_iter=5, features=features,
                                           gamma=10.0, alpha=0.1)
    assert model, "Invalid SSSRM instance!"

    # Create a Shared response S with K = 3
    theta = np.linspace(-4 * np.pi, 4 * np.pi, samples)
    z = np.linspace(-2, 2, samples)
    r = z**2 + 1
    x = r * np.sin(theta)
    y = r * np.cos(theta)

    S = np.vstack((x, y, z))
    S_align = S[:, :align_samples]
    S_classify = S[:, align_samples:]
    X = []
    Z = []
    Z2 = []
    W = []
    y = []
    Q, R = np.linalg.qr(np.random.random((voxels, features)))
    W.append(Q)
    X.append(Q.dot(S_align) + 0.1 * np.random.random((voxels, align_samples)))
    Z.append(Q.dot(S_classify)
             + 0.1 * np.random.random((voxels, samples - align_samples)))
    Z2.append(Q.dot(S_classify)
              + 0.1 * np.random.random((voxels, samples - align_samples)))
    y.append(np.repeat(
        np.arange(n_labels), (samples - align_samples)/n_labels))

    # Check that transform does NOT run before fitting the model
    with pytest.raises(NotFittedError):
        model.transform(X)
    print("Test: transforming before fitting the model")

    # Check that predict does NOT run before fitting the model
    with pytest.raises(NotFittedError):
        model.predict(X)
    print("Test: predicting before fitting the model")

    # Check that it does NOT run with 1 subject on X
    with pytest.raises(ValueError):
        model.fit(X, y, Z)
    print("Test: running SSSRM with 1 subject (alignment)")

    # Create more subjects align and classification data
    for subject in range(1, subjects):
        Q, R = np.linalg.qr(np.random.random((voxels, features)))
        W.append(Q)
        X.append(Q.dot(S_align)
                 + 0.1 * np.random.random((voxels, align_samples)))
        Z2.append(Q.dot(S_classify)
                  + 0.1 * np.random.random((voxels, samples - align_samples)))

    # Check that it does NOT run with 1 subject on y
    with pytest.raises(ValueError):
        model.fit(X, y, Z)
    print("Test: running SSSRM with 1 subject (labels)")

    # Create more subjects labels data
    for subject in range(1, subjects):
        y.append(np.repeat(
            np.arange(n_labels), (samples - align_samples)/n_labels))

    # Check that it does NOT run with 1 subject on Z
    with pytest.raises(ValueError):
        model.fit(X, y, Z)
    print("Test: running SSSRM with 1 subject (classif.)")

    # Check that alpha is in (0,1) range
    model_bad = brainiak.funcalign.sssrm.SSSRM(n_iter=1, features=features,
                                               gamma=10.0, alpha=1.5)
    assert model_bad, "Invalid SSSRM instance!"
    with pytest.raises(ValueError):
        model_bad.fit(X, y, Z)
    print("Test: running SSSRM with wrong alpha")

    # Check that gamma is positive
    model_bad = brainiak.funcalign.sssrm.SSSRM(n_iter=1, features=features,
                                               gamma=-0.1, alpha=0.2)
    assert model_bad, "Invalid SSSRM instance!"
    with pytest.raises(ValueError):
        model_bad.fit(X, y, Z)
    print("Test: running SSSRM with wrong gamma")


def test_sssrm():
    import numpy as np

    voxels = 100
    align_samples = 400
    samples = 500
    subjects = 2
    features = 3
    n_labels = 4

    model = brainiak.funcalign.sssrm.SSSRM(n_iter=5, features=features,
                                           gamma=10.0, alpha=0.1)
    assert model, "Invalid SSSRM instance!"

    # Create a Shared response S with K = 3
    theta = np.linspace(-4 * np.pi, 4 * np.pi, samples)
    z = np.linspace(-2, 2, samples)
    r = z**2 + 1
    x = r * np.sin(theta)
    y = r * np.cos(theta)

    S = np.vstack((x, y, z))
    S_align = S[:, :align_samples]
    S_classify = S[:, align_samples:]
    X = []
    Z = []
    Z2 = []
    y = []
    Q, R = np.linalg.qr(np.random.random((voxels, features)))
    X.append(Q.dot(S_align) + 0.1 * np.random.random((voxels, align_samples)))
    Z.append(Q.dot(S_classify)
             + 0.1 * np.random.random((voxels, samples - align_samples)))
    Z2.append(Q.dot(S_classify)
              + 0.1 * np.random.random((voxels, samples - align_samples)))
    y.append(np.repeat(
        np.arange(n_labels), (samples - align_samples)/n_labels))

    # Create more subjects align and classification data
    for subject in range(1, subjects):
        Q, R = np.linalg.qr(np.random.random((voxels, features)))
        X.append(Q.dot(S_align)
                 + 0.1 * np.random.random((voxels, align_samples)))
        Z2.append(Q.dot(S_classify)
                  + 0.1 * np.random.random((voxels, samples - align_samples)))

    # Create more subjects labels data
    for subject in range(1, subjects):
        y.append(np.repeat(
            np.arange(n_labels), (samples - align_samples)/n_labels))

    # Set the logging level to INFO
    import logging
    logging.basicConfig(level=logging.INFO)

    # Check that runs with 2 subject
    model.fit(X, y, Z2)
    print("Test: fitting SSSRM successfully")

    assert len(model.w_) == subjects, (
        "Invalid computation of SSSRM! (wrong # subjects in W)")
    for subject in range(subjects):
        assert model.w_[subject].shape[0] == voxels, (
            "Invalid computation of SSSRM! (wrong # voxels in W)")
        assert model.w_[subject].shape[1] == features, (
            "Invalid computation of SSSRM! (wrong # features in W)")
        ortho = np.linalg.norm(model.w_[subject].T.dot(model.w_[subject])
                               - np.eye(model.w_[subject].shape[1]),
                               'fro')
        assert ortho < 1e-7, "A Wi mapping is not orthonormal in SSSRM."
        difference = np.linalg.norm(X[subject]
                                    - model.w_[subject].dot(model.s_),
                                    'fro')
        datanorm = np.linalg.norm(X[subject], 'fro')
        assert difference/datanorm < 1.0, "Model seems incorrectly computed."
    assert model.s_.shape[0] == features, (
        "Invalid computation of SSSRM! (wrong # features in S)")
    assert model.s_.shape[1] == align_samples, (
        "Invalid computation of SSSRM! (wrong # samples in S)")

    # Check that it does run to compute the shared response after the model
    # computation
    new_s = model.transform(X)
    print("Test: transforming with SSSRM successfully")

    assert len(new_s) == subjects, (
        "Invalid computation of SSSRM! (wrong # subjects after transform)")
    for subject in range(subjects):
        assert new_s[subject].shape[0] == features, (
            "Invalid computation of SSSRM! (wrong # features after transform)")
        assert new_s[subject].shape[1] == align_samples, (
            "Invalid computation of SSSRM! (wrong # samples after transform)")

    # Check that it predicts with the model
    pred = model.predict(Z2)
    print("Test: predicting with SSSRM successfully")
    assert len(pred) == subjects, (
        "Invalid computation of SSSRM! (wrong # subjects after predict)")
    for subject in range(subjects):
        assert pred[subject].size == samples - align_samples, (
            "SSSRM: wrong # answers in predict")
        pred_labels = np.logical_and(pred[subject] >= 0,
                                     pred[subject] < n_labels)
        assert np.all(pred_labels), (
            "SSSRM: wrong class number output in predict")

    # Check that it does NOT run with non-matching number of subjects
    with pytest.raises(ValueError):
        model.transform(X[1])
    print("Test: transforming with non-matching number of subjects")

    # Check that it does not run without enough samples (TRs).
    with pytest.raises(ValueError):
        model.set_params(features=(align_samples + 1))
        model.fit(X, y, Z2)
    print("Test: not enough samples")

    # Check that it does not run with different number of samples (TRs)
    X2 = X
    X2[0] = Q.dot(S[:, :-2])
    with pytest.raises(ValueError):
        model.fit(X2, y, Z2)
    print("Test: different number of samples per subject")

    # Create one more subject
    Q, R = np.linalg.qr(np.random.random((voxels, features)))
    X.append(Q.dot(S_align) + 0.1 * np.random.random((voxels, align_samples)))
    Z2.append(Q.dot(S_classify)
              + 0.1 * np.random.random((voxels, samples - align_samples)))

    # Check that it does not run with different number of subjects in each
    # input
    with pytest.raises(ValueError):
        model.fit(X, y, Z2)
    print("Test: different number of subjects in the inputs")

    y.append(np.repeat(
        np.arange(n_labels), (samples - align_samples)/n_labels))
    with pytest.raises(ValueError):
        model.fit(X, y, Z)
    print("Test: different number of subjects in the inputs")



import os
import tempfile
import re

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal
from sklearn.exceptions import NotFittedError

from brainiak.funcalign.fastsrm import (
    FastSRM, _compute_and_save_corr_mat, _compute_and_save_subject_basis,
    _compute_basis_subject_online, _reduced_space_compute_shared_response,
    check_atlas, check_imgs, check_shared_response, create_temp_dir, fast_srm,
    reduce_data, safe_load)

from brainiak.funcalign.srm import DetSRM


def to_path(X, dirpath):
    """
    Save list of list of array to path and returns the path_like array
    Parameters
    ----------
    X: list of list of array
        input data
    dirpath: str
        dirpath
    Returns
    -------
    paths: array of str
        path arrays where all data are stored
    """
    paths = []
    for i, sessions in enumerate(X):
        sessions_path = []
        for j, session in enumerate(sessions):
            pth = "%i_%i" % (i, j)
            np.save(os.path.join(dirpath, pth), session)
            sessions_path.append(os.path.join(dirpath, pth + ".npy"))
        paths.append(sessions_path)
    return np.array(paths)


def generate_data(n_voxels,
                  n_timeframes,
                  n_subjects,
                  n_components,
                  datadir,
                  noise_level=0.1,
                  input_format="array"):
    n_sessions = len(n_timeframes)
    cumsum_timeframes = np.cumsum([0] + n_timeframes)
    slices_timeframes = [
        slice(cumsum_timeframes[i], cumsum_timeframes[i + 1])
        for i in range(n_sessions)
    ]

    # Create a Shared response S with K = 3
    theta = np.linspace(-4 * np.pi, 4 * np.pi, int(np.sum(n_timeframes)))
    z = np.linspace(-2, 2, int(np.sum(n_timeframes)))
    r = z**2 + 1
    x = r * np.sin(theta)
    y = r * np.cos(theta)

    S = np.vstack((x, y, z))

    # Generate fake data
    W = []
    X = []
    for subject in range(n_subjects):
        Q, R = np.linalg.qr(np.random.random((n_voxels, n_components)))
        W.append(Q.T)
        X_ = []
        for session in range(n_sessions):
            S_s = S[:, slices_timeframes[session]]
            S_s = S_s - np.mean(S_s, axis=1, keepdims=True)
            noise = noise_level * np.random.random(
                (n_voxels, n_timeframes[session]))
            noise = noise - np.mean(noise, axis=1, keepdims=True)
            data = Q.dot(S_s) + noise
            X_.append(data)
        X.append(X_)

    # create paths such that paths[i, j] contains data
    # of subject i during session j
    S = [(S[:, s] - np.mean(S[:, s], axis=1, keepdims=True))
         for s in slices_timeframes]

    if input_format == "array":
        paths = to_path(X, datadir)
        return paths, W, S

    elif input_format == "list_of_list":
        return X, W, S

    elif input_format == "list_of_array":
        return [
            np.concatenate([X[i][j].T for j in range(n_sessions)]).T
            for i in range(n_subjects)
        ], W, S
    else:
        raise ValueError("Wrong input_format")


def test_generated_data():
    with tempfile.TemporaryDirectory() as datadir:

        # We authorize different timeframes for different sessions
        # but they should be the same across subject
        n_voxels = 10
        n_timeframes = [25, 24]
        n_subjects = 2
        n_components = 3  # number of components used for SRM model
        n_sessions = len(n_timeframes)

        np.random.seed(0)
        paths, W, S = generate_data(n_voxels, n_timeframes, n_subjects,
                                    n_components, datadir)

        # Test if generated data has the good shape
        for subject in range(n_subjects):
            for session in range(n_sessions):
                assert (np.load(
                    paths[subject, session]).shape == (n_voxels,
                                                       n_timeframes[session]))

        # Test if generated basis have good shape
        assert len(W) == n_subjects
        for w in W:
            assert w.shape == (n_components, n_voxels)

        assert len(S) == n_sessions
        for j, s in enumerate(S):
            assert s.shape == (n_components, n_timeframes[j])


def test_bad_aggregate():
    with pytest.raises(ValueError,
                       match="aggregate can have only value mean or None"):
        FastSRM(aggregate="invalid")


def test_check_atlas():
    assert check_atlas(None) is None
    with pytest.raises(ValueError,
                       match=("Atlas is stored using type <class 'list'> "
                              "which is neither np.ndarray or str")):
        check_atlas([])

    A = np.random.rand(10, 100)
    assert check_atlas(A) == (10, 100)

    with tempfile.TemporaryDirectory() as datadir:
        f = os.path.join(datadir, "atlas")
        np.save(f, A)
        assert check_atlas(f + ".npy") == (10, 100)

    A = np.array([0, 0, 1, 1, 2, 2, 3, 3, 4, 5, 5])
    assert check_atlas(A) == (5, 11)

    with tempfile.TemporaryDirectory() as datadir:
        f = os.path.join(datadir, "atlas")
        np.save(f, A)
        assert check_atlas(f + ".npy") == (5, 11)

    with pytest.raises(ValueError,
                       match=("Atlas has 3 axes. It should have either "
                              "1 or 2 axes.")):
        check_atlas(np.random.rand(5, 1, 2))

    with pytest.raises(ValueError,
                       match=(r"Number of regions in the atlas is lower than "
                              r"the number of components \(3 < 5\)")):
        check_atlas(np.random.rand(3, 10), n_components=5)

    with pytest.raises(ValueError,
                       match=(r"Number of regions in the atlas is bigger than "
                              r"the number of voxels \(5 > 2\)")):
        check_atlas(np.random.rand(5, 2))


empty_list_error = "%s is a list of length 0 which is not valid"
array_type_error = ("%s should be of type np.ndarray but is of type %s")
array_2axis_error = ("%s must have exactly 2 axes but has %i axes")


def test_check_imgs():
    with pytest.raises(
            ValueError,
            match=(r"Since imgs is a list, it should be a list of list "
                   r"of arrays or a list "
                   r"of arrays but imgs\[0\] has type <class 'str'>")):
        check_imgs(["bla"])

    with pytest.raises(
            ValueError,
            match=("Input imgs should either be a list or an array but "
                   "has type <class 'str'>")):
        check_imgs("bla")

    with pytest.raises(ValueError, match=empty_list_error % "imgs"):
        check_imgs([])

    with pytest.raises(
            ValueError,
            match=r"imgs\[1\] has type <class 'str'> whereas imgs\[0\] has "
            "type <class 'int'>. This is inconsistent."):
        check_imgs([0, "bla"])

    with pytest.raises(ValueError, match=empty_list_error % r"imgs\[0\]"):
        check_imgs([[]])

    with pytest.raises(
            ValueError,
            match=(r"imgs\[1\] has length 1 whereas imgs\[0\] has length 2."
                   " All subjects should have the same number of sessions.")):
        check_imgs([["a", "a"], ["a"]])

    with pytest.raises(ValueError,
                       match=array_type_error %
                       (r"imgs\[0\]\[0\]", r"<class 'str'>")):
        check_imgs([["bka"]])

    with pytest.raises(ValueError,
                       match=array_2axis_error % (r"imgs\[0\]\[0\]", 1)):
        check_imgs([[np.random.rand(5)]])

    with pytest.raises(ValueError,
                       match=array_2axis_error % (r"imgs\[0\]", 1)):
        check_imgs([np.random.rand(5)])

    with pytest.raises(ValueError,
                       match=(r"imgs\[0, 0\] is stored using type "
                              "<class 'numpy.float64'> which is not a str")):
        check_imgs(np.random.rand(5, 3))

    with pytest.raises(ValueError, match=array_2axis_error % (r"imgs", 1)):
        check_imgs(np.random.rand(5))

    with pytest.raises(
            ValueError,
            match=("The number of subjects should be greater than 1")):
        check_imgs([np.random.rand(5, 3)])

    with pytest.raises(
            ValueError,
            match=("Subject 1 Session 0 does not have the same number "
                   "of timeframes as Subject 0 Session 0")):
        check_imgs([np.random.rand(10, 5), np.random.rand(10, 10)])

    with pytest.raises(
            ValueError,
            match=("Subject 1 Session 0 does not have the same number "
                   "of voxels as Subject 0 Session 0")):
        check_imgs([np.random.rand(10, 5), np.random.rand(20, 5)])

    with pytest.raises(
            ValueError,
            match=("Total number of timeframes is shorter than number "
                   r"of components \(5 < 8\)")):
        check_imgs([np.random.rand(10, 5),
                    np.random.rand(10, 5)],
                   n_components=8)

    with pytest.raises(
            ValueError,
            match=("Number of voxels in the atlas is not the same as "
                   r"the number of voxels in input data \(11 != 10\)")):
        check_imgs([np.random.rand(10, 5),
                    np.random.rand(10, 5)],
                   n_components=3,
                   atlas_shape=(8, 11))


def test_check_shared():
    n_subjects = 2
    n_sessions = 2
    input_shapes = np.zeros((n_subjects, n_sessions, 2))
    input_shapes[0, 0, 0] = 10
    input_shapes[0, 0, 1] = 3
    input_shapes[0, 1, 0] = 10
    input_shapes[0, 1, 1] = 2
    input_shapes[1, 0, 0] = 10
    input_shapes[1, 0, 1] = 3
    input_shapes[1, 1, 0] = 10
    input_shapes[1, 1, 1] = 2

    shared_list_list = [[
        np.array([[1, 2, 3], [4, 5, 6]]),
        np.array([[1, 2], [4, 5]]),
    ], [
        np.array([[2, 3, 4], [5, 6, 7]]),
        np.array([[2, 3], [5, 6]]),
    ]]

    shared_list_subjects = [
        np.array([[1, 2, 3, 1, 2], [4, 5, 6, 4, 5]]),
        np.array([[2, 3, 4, 2, 3], [5, 6, 7, 5, 6]])
    ]

    shared_list_sessions = [
        np.array([[1.5, 2.5, 3.5], [4.5, 5.5, 6.5]]),
        np.array([[1.5, 2.5], [4.5, 5.5]]),
    ]

    shared_array = np.array([[1.5, 2.5, 3.5, 1.5, 2.5],
                             [4.5, 5.5, 6.5, 4.5, 5.5]])

    with pytest.raises(ValueError,
                       match=(r"shared_response should be either a list or an "
                              "array but is of type <class 'str'>")):
        check_shared_response("bla")

    with pytest.raises(
            ValueError,
            match=(r"shared_response is a list but shared_response\[0\] "
                   "is neither a list or an array. This is invalid.")):
        check_shared_response(["bla", "bli"])

    with pytest.raises(
            ValueError,
            match=(r"shared_response\[0\] is a list but shared_response\[1\] "
                   "is not a list this is incompatible")):
        check_shared_response(
            [[np.random.rand(2, 2)], np.array([1])], aggregate=None)

    with pytest.raises(ValueError,
                       match=(r"shared_response\[1\] has len 1 whereas "
                              r"shared_response\[0\] has len 2. They should "
                              "have same len")):
        check_shared_response([[np.random.rand(2, 2),
                                np.random.rand(2, 2)], [np.random.rand(2, 2)]],
                              aggregate=None)

    with pytest.raises(
            ValueError,
            match=('Number of timeframes in input images during session 0 '
                   'does not match the number of timeframes during session '
                   r'0 of shared_response \(2 != 3\)')):
        check_shared_response(
            [np.random.rand(2, 2), np.random.rand(2, 2)],
            aggregate="mean",
            input_shapes=input_shapes)

    with pytest.raises(ValueError,
                       match=("Number of components in shared_response "
                              "during session 0 is different than "
                              "the number of components of the "
                              r"model \(2 != 4\)")):
        check_shared_response(np.random.rand(2, 10), n_components=4)

    with pytest.raises(ValueError,
                       match=("self.aggregate has value 'mean' but shared "
                              "response is a list of list. "
                              "This is incompatible")):
        added_session, reshaped_shared = check_shared_response(
            shared_list_list,
            aggregate="mean",
            n_components=2,
            input_shapes=input_shapes)

    added_session, reshaped_shared = check_shared_response(
        shared_list_subjects,
        aggregate=None,
        n_components=2,
        input_shapes=input_shapes)
    assert added_session
    assert_array_almost_equal(np.array(reshaped_shared),
                              shared_array.reshape(1, 2, 5))

    added_session, reshaped_shared = check_shared_response(
        shared_list_sessions,
        aggregate="mean",
        n_components=2,
        input_shapes=input_shapes)
    assert not added_session
    for j in range(len(reshaped_shared)):
        assert_array_almost_equal(reshaped_shared[j], shared_list_sessions[j])

    added_session, reshaped_shared = check_shared_response(
        shared_array,
        aggregate="mean",
        n_components=2,
        input_shapes=input_shapes)
    assert added_session
    assert_array_almost_equal(np.array(reshaped_shared),
                              shared_array.reshape(1, 2, 5))


def test_reduce_data_dummyatlases():
    n_jobs = 1
    with tempfile.TemporaryDirectory() as datadir:
        for n_timeframes in ([25, 24], [25, 25]):
            # We authorize different timeframes for different sessions
            # but they should be the same across subject
            n_voxels = 10
            n_subjects = 2
            n_components = 3  # number of components used for SRM model
            n_sessions = len(n_timeframes)

            np.random.seed(0)
            paths, _, _ = generate_data(n_voxels, n_timeframes, n_subjects,
                                        n_components, datadir)

            # test atlas that reduces nothing
            atlas = np.arange(1, n_voxels + 1)
            data = reduce_data(paths,
                               atlas=atlas,
                               n_jobs=n_jobs,
                               low_ram=False)
            for i in range(n_subjects):
                for j in range(n_sessions):
                    assert_array_almost_equal(data[i, j].T,
                                              np.load(paths[i, j]))

            # test atlas that reduces everything
            atlas = np.ones(n_voxels)
            data = reduce_data(paths,
                               atlas=atlas,
                               n_jobs=n_jobs,
                               low_ram=False)
            for i in range(n_subjects):
                for j in range(n_sessions):
                    assert_array_almost_equal(
                        data[i, j].T.flatten(),
                        np.mean(np.load(paths[i, j]), axis=0))


def test_reduce_data_outputshapes():
    n_jobs = 1
    with tempfile.TemporaryDirectory() as datadir:
        for n_timeframes in ([25, 24], [25, 25]):
            # We authorize different timeframes for different sessions
            # but they should be the same across subject
            n_voxels = 10
            n_subjects = 2
            n_components = 3  # number of components used for SRM model
            n_supervoxels = 5  # number of components of the atlas
            n_sessions = len(n_timeframes)

            np.random.seed(0)
            paths, _, _ = generate_data(n_voxels, n_timeframes, n_subjects,
                                        n_components, datadir)

            # Test if reduced data has the good shape
            # probabilistic atlas
            atlas = np.random.rand(n_supervoxels, n_voxels)
            data = reduce_data(paths,
                               atlas=atlas,
                               n_jobs=n_jobs,
                               low_ram=False,
                               temp_dir=None)

            for subject in range(n_subjects):
                for session in range(n_sessions):
                    assert data[subject, session].shape == (
                        n_timeframes[session], n_supervoxels)

            # deterministic atlas
            det_atlas = np.round(np.random.rand(n_voxels) * n_supervoxels)
            n_unique = len(np.unique(det_atlas)[1:])
            while n_unique != n_supervoxels:
                det_atlas = np.round(np.random.rand(n_voxels) * n_supervoxels)
                n_unique = len(np.unique(det_atlas)[1:])

            data = reduce_data(paths,
                               atlas=det_atlas,
                               n_jobs=n_jobs,
                               low_ram=True,
                               temp_dir=datadir)

            for subject in range(n_subjects):
                for session in range(n_sessions):
                    assert (np.load(data[subject, session]).shape == (
                        n_timeframes[session], n_supervoxels))


def test_reduced_data_srm():
    n_jobs = 1
    with tempfile.TemporaryDirectory() as datadir:
        np.random.seed(0)

        # We authorize different timeframes for different sessions but
        # they should be the same across subject
        n_voxels = 10
        n_timeframes = [25, 24]
        n_subjects = 5
        n_components = 3  # number of components used for SRM model
        n_sessions = len(n_timeframes)

        paths, W, S = generate_data(n_voxels, n_timeframes, n_subjects,
                                    n_components, datadir, 0)

        atlas = np.arange(1, n_voxels + 1)
        data = reduce_data(paths,
                           atlas=atlas,
                           n_jobs=n_jobs,
                           low_ram=False,
                           temp_dir=None)

        # Test if shared response has the good shape
        shared_response_list = \
            _reduced_space_compute_shared_response(
                data,
                reduced_basis_list=None,
                n_components=n_components
            )
        assert len(shared_response_list) == n_sessions

        for session in range(n_sessions):
            assert (shared_response_list[session].shape == (
                n_timeframes[session], n_components))

        # Test basis from shared response
        for i, sessions in enumerate(paths):
            basis = _compute_basis_subject_online(
                sessions, [S[k].T for k in range(len(S))])
            # test shape
            assert basis.shape == (n_components, n_voxels)
            # test orthogonality
            assert np.allclose(basis.dot(basis.T), np.eye(n_components))
            # test correctness
            assert_array_almost_equal(basis, W[i], 2)

        # Test reduced_data_shared_response
        shared_response_list = _reduced_space_compute_shared_response(
            data, reduced_basis_list=W, n_components=n_components)
        for session in range(n_sessions):
            S_real = np.mean(
                [data[i, session].dot(W[i].T) for i in range(n_subjects)],
                axis=0)
            assert_array_almost_equal(shared_response_list[session], S_real)
            assert_array_almost_equal(shared_response_list[session],
                                      S[session].T)

        shared_response_list = fast_srm(data, n_components=n_components)

        for i, sessions in enumerate(paths):
            basis = _compute_basis_subject_online(sessions,
                                                  shared_response_list)
            for j, session in enumerate(sessions):
                assert_array_almost_equal(shared_response_list[j].dot(basis),
                                          np.load(paths[i, j]).T, 3)


def test_compute_and_save():
    with tempfile.TemporaryDirectory() as datadir:
        np.random.seed(0)
        n_voxels = 10
        n_timeframes = [25, 24]
        n_subjects = 5
        n_components = 3  # number of components used for SRM model

        paths, W, S = generate_data(n_voxels, n_timeframes, n_subjects,
                                    n_components, datadir, 0)

        for m, subjects in enumerate(paths.T):
            for subject in subjects:
                _compute_and_save_corr_mat(subject, S[m].T, datadir)

        for i, sessions in enumerate(paths):
            basis = _compute_and_save_subject_basis(i, sessions, datadir)

            assert_array_almost_equal(np.load(basis), W[i], 3)


def test_fastsrm_class():
    n_jobs = 1
    with tempfile.TemporaryDirectory() as datadir:
        np.random.seed(0)

        # We authorize different timeframes for different sessions
        # but they should be the same across subject
        n_voxels = 10
        n_timeframes = [25, 24]
        n_subjects = 5
        n_components = 3  # number of components used for SRM model

        np.random.seed(0)
        paths, W, S = generate_data(n_voxels, n_timeframes, n_subjects,
                                    n_components, datadir, 0)

        atlas = np.arange(1, n_voxels + 1)
        srm = FastSRM(atlas=atlas,
                      n_components=n_components,
                      n_iter=10,
                      temp_dir=datadir,
                      low_ram=True,
                      verbose=True,
                      n_jobs=n_jobs)

        # Raises an error because model is not fitted yet
        with pytest.raises(NotFittedError):
            srm.transform(paths)

        srm.fit(paths)

        # An error can occur if temporary directory already exists
        with pytest.raises(ValueError,
                           match=("Path %s already exists. When a model "
                                  "is used, filesystem should be "
                                  r"cleaned by using the .clean\(\) "
                                  "method" % re.escape(srm.temp_dir))):
            # Error can occur if the filesystem is uncleaned
            create_temp_dir(srm.temp_dir)
            create_temp_dir(srm.temp_dir)

        shared_response = srm.transform(paths)

        # Raise error when wrong index
        with pytest.raises(ValueError,
                           match=("subjects_indexes should be either "
                                  "a list, an array or None but "
                                  "received type <class 'int'>")):
            srm.transform(paths, subjects_indexes=1000)

        with pytest.raises(ValueError,
                           match=("subjects_indexes should be either "
                                  "a list, an array or None but "
                                  "received type <class 'int'>")):
            srm.inverse_transform(shared_response, subjects_indexes=1000)

        with pytest.raises(ValueError,
                           match=("sessions_indexes should be either "
                                  "a list, an array or None but "
                                  "received type <class 'int'>")):
            srm.inverse_transform(shared_response, sessions_indexes=1000)

        with pytest.raises(ValueError,
                           match=("Input data imgs has len 5 whereas "
                                  "subject_indexes has len 1. "
                                  "The number of basis used to compute "
                                  "the shared response should be equal to "
                                  "the number of subjects in imgs")):
            srm.transform(paths, subjects_indexes=[0])

        with pytest.raises(ValueError,
                           match=("Index 1 of subjects_indexes has value 8 "
                                  "whereas value should be between 0 and 4")):
            srm.transform(paths[:2], subjects_indexes=[0, 8])

        with pytest.raises(ValueError,
                           match=("Index 1 of sessions_indexes has value 8 "
                                  "whereas value should be between 0 and 1")):
            srm.inverse_transform(shared_response, sessions_indexes=[0, 8])

        # Check behavior of .clean
        assert os.path.exists(srm.temp_dir)
        srm.clean()
        assert not os.path.exists(srm.temp_dir)


n_voxels = 10
n_subjects = 5
n_components = 3  # number of components used for SRM model


def apply_aggregate(shared_response, aggregate, input_format):
    if aggregate is None:
        if input_format == "list_of_array":
            return [np.mean(shared_response, axis=0)]
        else:
            return [
                np.mean([
                    shared_response[i][j] for i in range(len(shared_response))
                ],
                        axis=0) for j in range(len(shared_response[0]))
            ]
    else:
        if input_format == "list_of_array":
            return [shared_response]
        else:
            return shared_response


def apply_input_format(X, input_format):
    if input_format == "array":
        n_sessions = len(X[0])
        XX = [[np.load(X[i, j]) for j in range(len(X[i]))]
              for i in range(len(X))]
    elif input_format == "list_of_array":
        XX = [[x] for x in X]
        n_sessions = 1
    else:
        XX = X
        n_sessions = len(X[0])
    return XX, n_sessions


@pytest.mark.parametrize(
    "input_format, low_ram, tempdir, atlas, n_jobs, n_timeframes, aggregate",
    [("array", True, True, None, 1, [25, 25], "mean"),
     ("list_of_list", False, False, np.arange(1, n_voxels + 1), 1, [25, 24
                                                                    ], None),
     ("list_of_array", True, False, np.eye(n_voxels), 1, [25, 25], None),
     ("list_of_array", False, True, None, 1, [25, 24], "mean")])
def test_fastsrm_class_correctness(input_format, low_ram, tempdir, atlas,
                                   n_jobs, n_timeframes, aggregate):
    with tempfile.TemporaryDirectory() as datadir:
        np.random.seed(0)
        X, W, S = generate_data(n_voxels, n_timeframes, n_subjects,
                                n_components, datadir, 0, input_format)

        XX, n_sessions = apply_input_format(X, input_format)

        if tempdir:
            temp_dir = datadir
        else:
            temp_dir = None

        srm = FastSRM(atlas=atlas,
                      n_components=n_components,
                      n_iter=10,
                      temp_dir=temp_dir,
                      low_ram=low_ram,
                      verbose=True,
                      n_jobs=n_jobs,
                      aggregate=aggregate,
                      seed=0)

        # Check that there is no difference between fit_transform
        # and fit then transform

        srm.fit(X)
        basis = [safe_load(b) for b in srm.basis_list]
        shared_response_raw = srm.transform(X)
        shared_response = apply_aggregate(shared_response_raw, aggregate,
                                          input_format)
        shared_response_fittransform = apply_aggregate(srm.fit_transform(X),
                                                       aggregate, input_format)

        for j in range(n_sessions):
            assert_array_almost_equal(shared_response_fittransform[j],
                                      shared_response[j])

        # Check that the decomposition works
        for i in range(n_subjects):
            for j in range(n_sessions):
                assert_array_almost_equal(shared_response[j].T.dot(basis[i]),
                                          XX[i][j].T, 3)

        # Check that if we use all subjects but one if gives almost the
        # same shared response
        shared_response_partial_raw = srm.transform(X[1:5],
                                                    subjects_indexes=list(
                                                        range(1, 5)))

        shared_response_partial = apply_aggregate(shared_response_partial_raw,
                                                  aggregate, input_format)
        for j in range(n_sessions):
            assert_array_almost_equal(shared_response_partial[j],
                                      shared_response[j], 3)

        # Check that if we perform add 2 times the same subject we
        # obtain the same decomposition
        srm.add_subjects(X[:1], shared_response_raw)
        assert_array_almost_equal(safe_load(srm.basis_list[0]),
                                  safe_load(srm.basis_list[-1]))


@pytest.mark.parametrize(
    "input_format, low_ram, tempdir, atlas, n_jobs, n_timeframes, aggregate",
    [("array", True, True, None, 1, [25, 25], "mean"),
     ("list_of_list", False, False, np.arange(1, n_voxels + 1), 1, [25, 24
                                                                    ], None),
     ("list_of_array", True, False, np.eye(n_voxels), 1, [25, 25], None),
     ("list_of_array", False, True, None, 1, [25, 24], "mean")])
def test_class_srm_inverse_transform(input_format, low_ram, tempdir, atlas,
                                     n_jobs, n_timeframes, aggregate):

    with tempfile.TemporaryDirectory() as datadir:
        X, W, S = generate_data(n_voxels, n_timeframes, n_subjects,
                                n_components, datadir, 0, input_format)

        if tempdir:
            temp_dir = datadir
        else:
            temp_dir = None

        srm = FastSRM(atlas=atlas,
                      n_components=n_components,
                      n_iter=10,
                      temp_dir=temp_dir,
                      low_ram=low_ram,
                      verbose=True,
                      n_jobs=n_jobs,
                      aggregate=aggregate,
                      seed=0)

        # Check that there is no difference between fit_transform
        # and fit then transform

        srm.fit(X)
        shared_response_raw = srm.transform(X)
        # Check inverse transform
        if input_format == "list_of_array":
            reconstructed_data = srm.inverse_transform(shared_response_raw,
                                                       subjects_indexes=[0, 2])
            for i, ii in enumerate([0, 2]):
                assert_array_almost_equal(reconstructed_data[i], X[ii], 3)

            reconstructed_data = srm.inverse_transform(shared_response_raw,
                                                       subjects_indexes=None)
            for i in range(len(X)):
                assert_array_almost_equal(reconstructed_data[i], X[i], 3)
        else:
            reconstructed_data = srm.inverse_transform(shared_response_raw,
                                                       sessions_indexes=[1],
                                                       subjects_indexes=[0, 2])
            for i, ii in enumerate([0, 2]):
                for j, jj in enumerate([1]):
                    assert_array_almost_equal(reconstructed_data[i][j],
                                              safe_load(X[ii][jj]), 3)

            reconstructed_data = srm.inverse_transform(shared_response_raw,
                                                       subjects_indexes=None,
                                                       sessions_indexes=None)

            for i in range(len(X)):
                for j in range(len(X[i])):
                    assert_array_almost_equal(reconstructed_data[i][j],
                                              safe_load(X[i][j]), 3)


def test_fastsrm_identity():
    # In this function we test whether fastsrm and DetSRM have
    # identical behavior when atlas=None

    # We authorize different timeframes for different sessions
    # but they should be the same across subject
    n_voxels = 8
    n_timeframes = [4, 5, 6]
    n_subjects = 2
    n_components = 3  # number of components used for SRM model

    np.random.seed(0)
    paths, W, S = generate_data(n_voxels,
                                n_timeframes,
                                n_subjects,
                                n_components,
                                None,
                                input_format="list_of_array")

    # Test if generated data has the good shape
    for subject in range(n_subjects):
        assert paths[subject].shape == (n_voxels, np.sum([n_timeframes]))

    srm = DetSRM(n_iter=11, features=3, rand_seed=0)
    srm.fit(paths)
    shared = srm.transform(paths)

    fastsrm = FastSRM(atlas=None,
                      n_components=3,
                      verbose=True,
                      seed=0,
                      n_jobs=1,
                      n_iter=10)
    fastsrm.fit(paths)
    shared_fast = fastsrm.transform(paths)

    assert_array_almost_equal(shared_fast, np.mean(shared, axis=0))

    for i in range(n_subjects):
        assert_array_almost_equal(safe_load(fastsrm.basis_list[i]),
                                  srm.w_[i].T)


def load_and_concat(paths):
    """
    Take list of path and yields input data for ProbSRM
    Parameters
    ----------
    paths
    Returns
    -------
    X
    """
    X = []
    for i in range(len(paths)):
        X_i = np.concatenate(
            [np.load(paths[i, j]) for j in range(len(paths[i]))], axis=1)
        X.append(X_i)
    return X


def test_consistency_paths_data():
    with tempfile.TemporaryDirectory() as datadir:
        # In this function we test that input format
        # does not change the results

        n_voxels = 8
        n_timeframes = [4, 5, 6]
        n_subjects = 2
        n_components = 3  # number of components used for SRM model

        np.random.seed(0)
        paths, W, S = generate_data(n_voxels,
                                    n_timeframes,
                                    n_subjects,
                                    n_components,
                                    datadir,
                                    input_format="array")

        print()
        print("shape", paths.shape)

        fastsrm = FastSRM(
            n_components=3,
            atlas=None,
            verbose=True,
            seed=0,
            n_jobs=1,
            n_iter=10,
        )

        fastsrm.fit(paths)
        b0 = fastsrm.basis_list[0]

        fastsrm.fit(load_and_concat(paths))
        b1 = fastsrm.basis_list[0]

        assert_array_almost_equal(b0, b1)



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from sklearn.exceptions import NotFittedError
import pytest
from mpi4py import MPI


@pytest.mark.mpiexec(n=2)
def test_distributed_srm():  # noqa: C901
    import brainiak.funcalign.srm
    s = brainiak.funcalign.srm.SRM()
    assert s, "Invalid SRM instance!"

    import numpy as np
    np.random.seed(0)
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    nrank = comm.Get_size()

    voxels = 100
    samples = 500
    subjects = 2
    features = 3

    s = brainiak.funcalign.srm.SRM(n_iter=5, features=features, comm=comm)
    assert s, "Invalid SRM instance!"

    # Create a Shared response S with K = 3
    theta = np.linspace(-4 * np.pi, 4 * np.pi, samples)
    z = np.linspace(-2, 2, samples)
    r = z**2 + 1
    x = r * np.sin(theta)
    y = r * np.cos(theta)

    S = np.vstack((x, y, z))

    # DSRM: broadcast S
    S = comm.bcast(S)

    X = []
    W = []
    # DSRM: only append on rank 0
    Q, R = np.linalg.qr(np.random.random((voxels, features)))
    tmp_noise = 0.1*np.random.random((voxels, samples))
    if rank == 0:
        W.append(Q)
        X.append(Q.dot(S) + tmp_noise)
    else:
        W.append(None)
        X.append(None)

    # Check that transform does NOT run before fitting the model
    with pytest.raises(NotFittedError):
        s.transform(X)
    if rank == 0:
        print("Test: transforming before fitting the model")

    # Check that it does NOT run with 1 subject
    with pytest.raises(ValueError):
        s.fit(X)
    if rank == 0:
        print("Test: running SRM with 1 subject")

    # DSRM: cyclic distribution of subject data, otherwise None
    for subject in range(1, subjects):
        Q, R = np.linalg.qr(np.random.random((voxels, features)))
        tmp_noise = 0.1*np.random.random((voxels, samples))
        if subject % nrank == rank:
            W.append(Q)
            X.append(Q.dot(S) + tmp_noise)
        else:
            W.append(None)
            X.append(None)

    # Check that runs with 2 subject
    s.fit(X)
    from pathlib import Path
    sr_v0_4 = np.load(Path(__file__).parent / "sr_v0_4.npz")['sr']
    assert (np.allclose(sr_v0_4, s.s_))

    assert len(s.w_) == subjects, (
        "Invalid computation of SRM! (wrong # subjects in W)")
    for subject in range(subjects):
        if s.w_[subject] is not None:
            assert s.w_[subject].shape[0] == voxels, (
                "Invalid computation of SRM! (wrong # voxels in W)")
            assert s.w_[subject].shape[1] == features, (
                "Invalid computation of SRM! (wrong # features in W)")
            ortho = np.linalg.norm(s.w_[subject].T.dot(s.w_[subject])
                                   - np.eye(s.w_[subject].shape[1]),
                                   'fro')
            assert ortho < 1e-7, "A Wi mapping is not orthonormal in SRM."
            difference = np.linalg.norm(X[subject] - s.w_[subject].dot(s.s_),
                                        'fro')
            datanorm = np.linalg.norm(X[subject], 'fro')
            assert difference/datanorm < 1.0, (
                "Model seems incorrectly computed.")

    assert s.s_.shape[0] == features, (
        "Invalid computation of SRM! (wrong # features in S)")
    assert s.s_.shape[1] == samples, (
        "Invalid computation of SRM! (wrong # samples in S)")

    # Check that it does run to compute the shared response after the model
    # computation
    new_s = s.transform(X)

    assert len(new_s) == subjects, (
        "Invalid computation of SRM! (wrong # subjects after transform)")
    for subject in range(subjects):
        if new_s[subject] is not None:
            assert new_s[subject].shape[0] == features, (
                "Invalid computation of SRM! (wrong # features after "
                "transform)")
            assert new_s[subject].shape[1] == samples, (
                "Invalid computation of SRM! (wrong # samples after "
                "transform)")

    # Check that it does NOT run with non-matching number of subjects
    with pytest.raises(ValueError):
        s.transform([X[1]])
    if rank == 0:
        print("Test: transforming with non-matching number of subjects")

    # Check that it does not run without enough samples (TRs).
    with pytest.raises(ValueError):
        s.set_params(features=(samples+1))
        s.fit(X)
    if rank == 0:
        print("Test: not enough samples")

    # Check that it does not run with different number of samples (TRs)
    if rank == 0:
        S2 = S[:, :-2]
        X.append(Q.dot(S2))
    else:
        X.append(None)

    with pytest.raises(ValueError):
        s.fit(X)
    if rank == 0:
        print("Test: different number of samples per subject")



from brainiak.eventseg.event import EventSegment
from scipy.special import comb
import numpy as np
import pytest
from sklearn.exceptions import NotFittedError


def test_create_event_segmentation():
    es = EventSegment(5)
    assert es, "Invalid EventSegment instance"


def test_fit_shapes():
    K = 5
    V = 3
    T = 10
    es = EventSegment(K, n_iter=2)
    sample_data = np.random.rand(V, T)
    es.fit(sample_data.T)

    assert es.segments_[0].shape == (T, K), "Segmentation from fit " \
                                            "has incorrect shape"
    assert np.isclose(np.sum(es.segments_[0], axis=1), np.ones(T)).all(), \
        "Segmentation from learn_events not correctly normalized"

    T2 = 15
    sample_data2 = np.random.rand(V, T2)
    test_segments, test_ll = es.find_events(sample_data2.T)

    assert test_segments.shape == (T2, K), "Segmentation from find_events " \
                                           "has incorrect shape"
    assert np.isclose(np.sum(test_segments, axis=1), np.ones(T2)).all(), \
        "Segmentation from find_events not correctly normalized"

    es_invalid = EventSegment(K)
    with pytest.raises(ValueError):
        es_invalid.model_prior(K-1)
        # ``with`` block is about to end with no error.
        pytest.fail("T < K should cause error")
    with pytest.raises(ValueError):
        es_invalid.set_event_patterns(np.zeros((V, K-1)))
        pytest.fail("#Events < K should cause error")


def test_simple_boundary():
    es = EventSegment(2)
    random_state = np.random.RandomState(0)

    sample_data = np.array([[1, 1, 1, 0, 0, 0, 0], [0, 0, 0, 1, 1, 1, 1]]) + \
        random_state.rand(2, 7) * 10
    es.fit(sample_data.T)

    events = np.argmax(es.segments_[0], axis=1)
    assert np.array_equal(events, [0, 0, 0, 1, 1, 1, 1]),\
        "Failed to correctly segment two events"

    events_predict = es.predict(sample_data.T)
    assert np.array_equal(events_predict, [0, 0, 0, 1, 1, 1, 1]), \
        "Error in predict interface"


def test_event_transfer():
    es = EventSegment(2)
    sample_data = np.asarray([[1, 1, 1, 0, 0, 0, 0], [0, 0, 0, 1, 1, 1, 1]])

    with pytest.raises(NotFittedError):
        seg = es.find_events(sample_data.T)[0]
        pytest.fail("Should need to set variance")

    with pytest.raises(NotFittedError):
        seg = es.find_events(sample_data.T, np.asarray([1, 1]))[0]
        pytest.fail("Should need to set patterns")

    es.set_event_patterns(np.asarray([[1, 0], [0, 1]]))
    seg = es.find_events(sample_data.T, np.asarray([1, 1]))[0]

    events = np.argmax(seg, axis=1)
    assert np.array_equal(events, [0, 0, 0, 1, 1, 1, 1]),\
        "Failed to correctly transfer two events to new data"


def test_weighted_var():
    es = EventSegment(2)

    D = np.zeros((8, 4))
    for t in range(4):
        D[t, :] = (1/np.sqrt(4/3)) * np.array([-1, -1, 1, 1])
    for t in range(4, 8):
        D[t, :] = (1 / np.sqrt(4 / 3)) * np.array([1, 1, -1, -1])
    mean_pat = D[[0, 4], :].T

    weights = np.zeros((8, 2))
    weights[:, 0] = [1, 1, 1, 1, 0, 0, 0, 0]
    weights[:, 1] = [0, 0, 0, 0, 1, 1, 1, 1]
    assert np.array_equal(
        es.calc_weighted_event_var(D, weights, mean_pat), [0, 0]),\
        "Failed to compute variance with 0/1 weights"

    weights[:, 0] = [1, 1, 1, 1, 0.5, 0.5, 0.5, 0.5]
    weights[:, 1] = [0.5, 0.5, 0.5, 0.5, 1, 1, 1, 1]
    true_var = (4 * 0.5 * 12)/(6 - 5/6) * np.ones(2) / 4
    assert np.allclose(
        es.calc_weighted_event_var(D, weights, mean_pat), true_var),\
        "Failed to compute variance with fractional weights"


def test_sym():
    es = EventSegment(4)

    evpat = np.repeat(np.arange(10).reshape(-1, 1), 4, axis=1)
    es.set_event_patterns(evpat)

    D = np.repeat(np.arange(10).reshape(1, -1), 20, axis=0)
    ev = es.find_events(D, var=1)[0]

    # Check that events 1-4 and 2-3 are symmetric
    assert np.all(np.isclose(ev[:, :2], np.fliplr(np.flipud(ev[:, 2:])))),\
        "Fit with constant data is not symmetric"


def test_chains():
    es = EventSegment(5, event_chains=np.array(['A', 'A', 'B', 'B', 'B']))
    sample_data = np.array([[0, 0, 0], [1, 1, 1]])

    with pytest.raises(RuntimeError):
        seg = es.fit(sample_data.T)[0]
        pytest.fail("Can't use fit() with event chains")

    es.set_event_patterns(np.array([[1, 1, 0, 0, 0],
                                    [0, 0, 1, 1, 1]]))
    seg = es.find_events(sample_data.T, 0.1)[0]

    ev = np.nonzero(seg > 0.99)[1]
    assert np.array_equal(ev, [2, 3, 4]),\
        "Failed to fit with multiple chains"


def test_prior():
    K = 10
    T = 100

    es = EventSegment(K)
    mp = es.model_prior(T)[0]

    p_bound = np.zeros((T, K-1))
    norm = comb(T-1, K-1)
    for t in range(T-1):
        for k in range(K-1):
            # See supplementary material of Neuron paper
            # https://doi.org/10.1016/j.neuron.2017.06.041
            p_bound[t+1, k] = comb(t, k) * comb(T-t-2, K-k-2) / norm
    p_bound = np.cumsum(p_bound, axis=0)

    mp_gt = np.zeros((T, K))
    for k in range(K):
        if k == 0:
            mp_gt[:, k] = 1 - p_bound[:, 0]
        elif k == K - 1:
            mp_gt[:, k] = p_bound[:, k-1]
        else:
            mp_gt[:, k] = p_bound[:, k-1] - p_bound[:, k]

    assert np.all(np.isclose(mp, mp_gt)),\
        "Prior does not match analytic solution"


def test_split_merge():
    ev = np.array(
        [0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
         3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4])
    random_state = np.random.RandomState(0)
    ev_pat = random_state.rand(5, 10)
    D = np.zeros((len(ev), 10))
    for t in range(len(ev)):
        D[t, :] = ev_pat[ev[t], :] + 0.1*random_state.rand(10)

    hmm_sm = EventSegment(5, split_merge=True, split_merge_proposals=2)
    hmm_sm.fit(D)
    hmm_events = np.argmax(hmm_sm.segments_[0], axis=1)

    assert np.all(ev == hmm_events),\
        "Merge/split fails to find highly uneven events"


def test_sym_ll():
    ev = np.array([0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2])
    random_state = np.random.RandomState(0)
    ev_pat = random_state.rand(3, 10)
    D_forward = np.zeros((len(ev), 10))
    for t in range(len(ev)):
        D_forward[t, :] = ev_pat[ev[t], :] + 0.1 * random_state.rand(10)
    D_backward = np.flip(D_forward, axis=0)

    hmm_forward = EventSegment(3)
    hmm_forward.set_event_patterns(ev_pat.T)
    _, ll_forward = hmm_forward.find_events(D_forward, var=1)

    hmm_backward = EventSegment(3)
    hmm_backward.set_event_patterns(np.flip(ev_pat.T, axis=1))
    _, ll_backward = hmm_backward.find_events(D_backward, var=1)

    assert np.allclose(ll_forward, ll_backward),\
        "Log-likelihood not symmetric forward/backward"



#  Copyright 2017 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Iterable, Sequence

import numpy as np
import pytest

from nibabel.nifti1 import Nifti1Pair
from nibabel.spatialimages import SpatialImage

from brainiak.image import (mask_image, mask_images, MaskedMultiSubjectData,
                            multimask_images, SingleConditionSpec)


@pytest.fixture
def masked_multi_subject_data(masked_images):
    return np.stack(masked_images, axis=-1)


class TestMaskedMultiSubjectData:
    def test_from_masked_images(self, masked_images,
                                masked_multi_subject_data):
        result = MaskedMultiSubjectData.from_masked_images(masked_images,
                                                           len(masked_images))
        assert np.array_equal(np.moveaxis(result, 1, 0),
                              masked_multi_subject_data)


@pytest.fixture
def condition_spec() -> SingleConditionSpec:
    return np.array([[[1, 1, 1, 1, 0, 0, 0, 0, 0, 0],
                      [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]],
                     [[0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                      [0, 0, 0, 0, 0, 1, 1, 1, 1, 0]]],
                    dtype=np.int8).view(SingleConditionSpec)


class TestUniqueLabelConditionSpec:
    def test_extract_labels(self, condition_spec: SingleConditionSpec
                            ) -> None:
        assert np.array_equal(condition_spec.extract_labels(),
                              np.array([0, 1]))


@pytest.fixture
def spatial_image() -> SpatialImage:
    return Nifti1Pair(
        np.array(
            [
                [
                    [[0, 0, 0, 0],
                     [0, 0, 0, 0],
                     [0, 0, 0, 0],
                     [0, 0, 0, 0]],
                    [[0, 0, 0, 0],
                     [0, 1, 0, 0],
                     [0, 0, 1, 0],
                     [0, 0, 0, 0]],
                    [[0, 0, 0, 0],
                     [0, 0, 1, 0],
                     [0, 1, 0, 0],
                     [0, 0, 0, 0]],
                    [[0, 0, 0, 0],
                     [0, 0, 0, 0],
                     [0, 0, 0, 0],
                     [0, 0, 0, 0]],
                ]
            ],
            dtype=float,
        ).reshape(4, 4, 4, 1),
        np.eye(4),
    )


@pytest.fixture
def mask() -> np.ndarray:
    return np.array([[[0, 0, 0, 0],
                      [0, 0, 0, 0],
                      [0, 0, 0, 0],
                      [0, 0, 0, 0]],
                     [[0, 0, 0, 0],
                      [0, 1, 1, 0],
                      [0, 1, 1, 0],
                      [0, 0, 0, 0]],
                     [[0, 0, 0, 0],
                      [0, 1, 1, 0],
                      [0, 1, 1, 0],
                      [0, 0, 0, 0]],
                     [[0, 0, 0, 0],
                      [0, 0, 0, 0],
                      [0, 0, 0, 0],
                      [0, 0, 0, 0]]], dtype=bool)


@pytest.fixture
def masked_data() -> np.ndarray:
    return np.array([[1, 0, 0, 1, 0, 1, 1, 0]]).reshape(8, 1)


@pytest.fixture
def images(spatial_image: SpatialImage) -> Iterable[SpatialImage]:
    images = [spatial_image]
    image_data = spatial_image.get_fdata().copy()
    image_data[1, 1, 1, 0] = 2
    images.append(Nifti1Pair(image_data, np.eye(4)))
    return images


@pytest.fixture
def masks(mask: np.ndarray) -> Sequence[np.ndarray]:
    masks = [mask]
    mask2 = mask.copy()
    mask2[0, 0, 0] = 1
    masks.append(mask2)
    mask3 = mask.copy()
    mask3[2, 2, 2] = 0
    masks.append(mask3)
    return masks


@pytest.fixture
def multimasked_images(masked_data) -> Iterable[Iterable[np.ndarray]]:
    masked_data_2 = np.concatenate(([[2]], masked_data[1:, :]))
    return [[masked_data, np.concatenate(([[0]], masked_data)),
             masked_data[:-1, :]],
            [masked_data_2, np.concatenate((np.array([[0]]), masked_data_2)),
             masked_data_2[:-1, :]]]


@pytest.fixture
def masked_images(multimasked_images) -> Iterable[np.ndarray]:
    return [multimasked_image[0] for multimasked_image in multimasked_images]


def test_mask_image(spatial_image: SpatialImage, mask: np.ndarray,
                    masked_data: np.ndarray) -> None:
    result = mask_image(spatial_image, mask)
    assert np.array_equal(result, masked_data)


def test_mask_image_with_type(spatial_image: SpatialImage, mask: np.ndarray,
                              masked_data: np.ndarray) -> None:
    masked_data_type = np.float32
    result = mask_image(spatial_image, mask, masked_data_type)
    assert result.dtype == masked_data_type
    assert np.allclose(result, masked_data)


def test_multimask_images(
        images: Iterable[SpatialImage],
        masks: Sequence[np.ndarray],
        multimasked_images: Iterable[Iterable[np.ndarray]]
        ) -> None:
    result = multimask_images(images, masks)
    for result_images, expected_images in zip(result,
                                              multimasked_images):
        for result_image, expected_image in zip(result_images,
                                                expected_images):
            assert np.array_equal(result_image, expected_image)


def test_mask_images(
        images: Iterable[SpatialImage],
        mask: np.ndarray,
        masked_images: Iterable[np.ndarray]
        ) -> None:
    result = mask_images(images, mask)
    for result_image, expected_image in zip(result, masked_images):
        assert np.array_equal(result_image, expected_image)



#  Copyright 2016 Mingbo Cai, Princeton Neuroscience Instititute,
#  Princeton University
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import pytest


def test_can_instantiate():
    import brainiak.reprsimil.brsa
    s = brainiak.reprsimil.brsa.BRSA()
    assert s, "Invalid BRSA instance!"

    s = brainiak.reprsimil.brsa.BRSA(
        n_iter=50, rank=5, auto_nuisance=False, n_nureg=2, nureg_method='ICA',
        baseline_single=False, init_iter=5, GP_space=True, GP_inten=True,
        tol=2e-3, eta=0.001, space_smooth_range=10.0, inten_smooth_range=100.0,
        tau_range=2.0,
        tau2_prior=brainiak.reprsimil.brsa.prior_GP_var_inv_gamma,
        optimizer='CG', random_state=100, anneal_speed=20)
    assert s, "Invalid BRSA instance!"


def test_fit():
    from brainiak.reprsimil.brsa import BRSA
    import brainiak.utils.utils as utils
    import scipy.stats
    import numpy as np
    import os.path
    np.random.seed(10)
    file_path = os.path.join(os.path.dirname(__file__), "example_design.1D")
    # Load an example design matrix
    design = utils.ReadDesign(fname=file_path)

    # concatenate it by 2 times, mimicking 2 runs of itenditcal timing
    n_run = 2
    design.design_task = np.tile(design.design_task[:, :-1], [n_run, 1])
    design.n_TR = design.n_TR * n_run

    # start simulating some data
    n_V = 50
    n_C = np.size(design.design_task, axis=1)
    n_T = design.n_TR

    noise_bot = 0.5
    noise_top = 5.0
    noise_level = np.random.rand(n_V) * (noise_top - noise_bot) + noise_bot
    # noise level is random.

    # AR(1) coefficient
    rho1_top = 0.8
    rho1_bot = -0.2
    rho1 = np.random.rand(n_V) * (rho1_top - rho1_bot) + rho1_bot

    # generating noise
    noise = np.zeros([n_T, n_V])
    noise[0, :] = np.random.randn(n_V) * noise_level / np.sqrt(1 - rho1**2)
    for i_t in range(1, n_T):
        noise[i_t, :] = noise[i_t - 1, :] * rho1 + \
            np.random.randn(n_V) * noise_level

    # ideal covariance matrix
    ideal_cov = np.zeros([n_C, n_C])
    ideal_cov = np.eye(n_C) * 0.6
    ideal_cov[0:4, 0:4] = 0.2
    for cond in range(0, 4):
        ideal_cov[cond, cond] = 2
    ideal_cov[5:9, 5:9] = 0.9
    for cond in range(5, 9):
        ideal_cov[cond, cond] = 1
    L_full = np.linalg.cholesky(ideal_cov)

    # generating signal
    snr_level = 5.0  # test with high SNR
    # snr = np.random.rand(n_V)*(snr_top-snr_bot)+snr_bot
    # Notice that accurately speaking this is not snr. the magnitude of signal
    # depends
    # not only on beta but also on x.
    inten = np.random.rand(n_V) * 20.0

    # parameters of Gaussian process to generate pseuso SNR
    tau = 1.0
    smooth_width = 5.0
    inten_kernel = 1.0

    coords = np.arange(0, n_V)[:, None]

    dist2 = np.square(coords - coords.T)

    inten_tile = np.tile(inten, [n_V, 1])
    inten_diff2 = (inten_tile - inten_tile.T)**2

    K = np.exp(-dist2 / smooth_width**2 / 2.0 - inten_diff2 /
               inten_kernel**2 / 2.0) * tau**2 + np.eye(n_V) * tau**2 * 0.001

    L = np.linalg.cholesky(K)
    snr = np.exp(np.dot(L, np.random.randn(n_V))) * snr_level
    sqrt_v = noise_level * snr
    betas_simulated = np.dot(L_full, np.random.randn(n_C, n_V)) * sqrt_v
    signal = np.dot(design.design_task, betas_simulated)

    # Adding noise to signal as data
    Y = signal + noise + inten

    scan_onsets = np.linspace(0, design.n_TR, num=n_run + 1)

    # Test fitting with GP prior.
    brsa = BRSA(GP_space=True, GP_inten=True, n_iter=5,
                init_iter=10, auto_nuisance=False, tol=2e-3)

    # We also test that it can detect baseline regressor included in the
    # design matrix for task conditions
    wrong_design = np.insert(design.design_task, 0, 1, axis=1)
    with pytest.raises(ValueError) as excinfo:
        brsa.fit(X=Y, design=wrong_design, scan_onsets=scan_onsets,
                 coords=coords, inten=inten)
    assert ('Your design matrix appears to have included baseline time series.'
            in str(excinfo.value))
    # Now we fit with the correct design matrix.
    brsa.fit(X=Y, design=design.design_task, scan_onsets=scan_onsets,
             coords=coords, inten=inten)

    # Check that result is significantly correlated with the ideal covariance
    # matrix
    u_b = brsa.U_
    u_i = ideal_cov
    p = scipy.stats.spearmanr(u_b[np.tril_indices_from(u_b)],
                              u_i[np.tril_indices_from(u_i)])[1]
    assert p < 0.01, (
        "Fitted covariance matrix does not correlate with ideal covariance "
        "matrix!")
    # check that the recovered SNRs makes sense
    p = scipy.stats.pearsonr(brsa.nSNR_, snr)[1]
    assert p < 0.01, "Fitted SNR does not correlate with simulated SNR!"
    assert np.isclose(np.mean(np.log(brsa.nSNR_)), 0), "nSNR_ not normalized!"
    p = scipy.stats.pearsonr(brsa.sigma_, noise_level)[1]
    assert p < 0.01, (
        "Fitted noise level does not correlate with simulated noise level!")
    p = scipy.stats.pearsonr(brsa.rho_, rho1)[1]
    assert p < 0.01, (
        "Fitted AR(1) coefficient does not correlate with simulated values!")

    noise_new = np.zeros([n_T, n_V])
    noise_new[0, :] = np.random.randn(n_V) * noise_level / np.sqrt(1 - rho1**2)
    for i_t in range(1, n_T):
        noise_new[i_t, :] = noise_new[i_t - 1, :] * \
            rho1 + np.random.randn(n_V) * noise_level

    Y_new = signal + noise_new + inten
    ts, ts0 = brsa.transform(Y_new, scan_onsets=scan_onsets)
    p = scipy.stats.pearsonr(ts[:, 0], design.design_task[:, 0])[1]
    assert p < 0.01, (
        "Recovered time series does not correlate with true time series!")
    assert np.shape(ts) == (n_T, n_C) and np.shape(ts0) == (n_T, 1), (
        "Wrong shape in returned time series by transform function!")

    [score, score_null] = brsa.score(
        X=Y_new, design=design.design_task, scan_onsets=scan_onsets)
    assert score > score_null, (
        "Full model does not win over null model on data containing signal")

    [score, score_null] = brsa.score(X=noise_new + inten,
                                     design=design.design_task,
                                     scan_onsets=scan_onsets)
    assert score < score_null, (
        "Null model does not win over full model on data without signal")

    # Test fitting with lower rank, nuisance regressors and without GP prior
    rank = n_C - 1
    n_nureg = 1
    brsa = BRSA(rank=rank, n_nureg=n_nureg, tol=2e-3,
                n_iter=8, init_iter=4, auto_nuisance=True)
    brsa.fit(X=Y, design=design.design_task, scan_onsets=scan_onsets)
    # u_b = brsa.U_
    u_i = ideal_cov
    p = scipy.stats.spearmanr(u_b[np.tril_indices_from(u_b)], u_i[
                              np.tril_indices_from(u_i)])[1]
    assert p < 0.01, (
        "Fitted covariance matrix does not correlate with ideal covariance "
        "matrix!")
    # check that the recovered SNRs makes sense
    p = scipy.stats.pearsonr(brsa.nSNR_, snr)[1]
    assert p < 0.01, "Fitted SNR does not correlate with simulated SNR!"
    assert np.isclose(np.mean(np.log(brsa.nSNR_)), 0), "nSNR_ not normalized!"
    p = scipy.stats.pearsonr(brsa.sigma_, noise_level)[1]
    assert p < 0.01, (
        "Fitted noise level does not correlate with simulated noise level!")
    p = scipy.stats.pearsonr(brsa.rho_, rho1)[1]
    assert p < 0.01, (
        "Fitted AR(1) coefficient does not correlate with simulated values!")

    assert (not hasattr(brsa, 'bGP_')
            and not hasattr(brsa, 'lGPspace_')
            and not hasattr(brsa, 'lGPinten_')
            ), ("the BRSA object should not have parameters of GP if GP is "
                "not requested.")
    # GP parameters are not set if not requested
    assert brsa.beta0_.shape[0] == n_nureg + 1, 'Shape of beta0 incorrect'
    p = scipy.stats.pearsonr(brsa.beta0_[0, :], inten)[1]
    assert p < 0.01, (
        'recovered beta0 does not correlate with the baseline of voxels.')
    assert np.shape(brsa.L_) == (
        n_C, rank), 'Cholesky factor should have shape of (n_C, rank)'

    # Test fitting with GP over just spatial coordinates.
    brsa = BRSA(GP_space=True, baseline_single=False,
                tol=2e-3, n_iter=4, init_iter=4)
    brsa.fit(X=Y, design=design.design_task,
             scan_onsets=scan_onsets, coords=coords)
    # Check that result is significantly correlated with the ideal covariance
    # matrix
    u_b = brsa.U_
    u_i = ideal_cov
    p = scipy.stats.spearmanr(u_b[np.tril_indices_from(u_b)], u_i[
                              np.tril_indices_from(u_i)])[1]
    assert p < 0.01, (
        "Fitted covariance matrix does not correlate with ideal covariance "
        "matrix!")
    # check that the recovered SNRs makes sense
    p = scipy.stats.pearsonr(brsa.nSNR_, snr)[1]
    assert p < 0.01, "Fitted SNR does not correlate with simulated SNR!"
    assert np.isclose(np.mean(np.log(brsa.nSNR_)), 0), "nSNR_ not normalized!"
    p = scipy.stats.pearsonr(brsa.sigma_, noise_level)[1]
    assert p < 0.01, (
        "Fitted noise level does not correlate with simulated noise level!")
    p = scipy.stats.pearsonr(brsa.rho_, rho1)[1]
    assert p < 0.01, (
        "Fitted AR(1) coefficient does not correlate with simulated values!")
    assert not hasattr(brsa, 'lGPinten_'), (
        "the BRSA object should not have parameters of lGPinten_ if only "
        "smoothness in space is requested.")
    # GP parameters are not set if not requested


def test_gradient():
    from brainiak.reprsimil.brsa import BRSA
    import brainiak.utils.utils as utils
    import numpy as np
    import os.path
    import numdifftools as nd

    np.random.seed(100)
    file_path = os.path.join(os.path.dirname(__file__), "example_design.1D")
    # Load an example design matrix
    design = utils.ReadDesign(fname=file_path)
    n_run = 4
    # concatenate it by 4 times, mimicking 4 runs of itenditcal timing
    design.design_task = np.tile(design.design_task[:, :-1], [n_run, 1])
    design.n_TR = design.n_TR * n_run

    # start simulating some data
    n_V = 30
    n_C = np.size(design.design_task, axis=1)
    n_T = design.n_TR

    noise_bot = 0.5
    noise_top = 1.5
    noise_level = np.random.rand(n_V) * (noise_top - noise_bot) + noise_bot
    # noise level is random.

    # AR(1) coefficient
    rho1_top = 0.8
    rho1_bot = -0.2
    rho1 = np.random.rand(n_V) * (rho1_top - rho1_bot) + rho1_bot

    # generating noise
    noise = np.zeros([n_T, n_V])
    noise[0, :] = np.random.randn(n_V) * noise_level / np.sqrt(1 - rho1**2)
    for i_t in range(1, n_T):
        noise[i_t, :] = noise[i_t - 1, :] * rho1 + \
            np.random.randn(n_V) * noise_level

    # ideal covariance matrix
    ideal_cov = np.zeros([n_C, n_C])
    ideal_cov = np.eye(n_C) * 0.6
    ideal_cov[0, 0] = 0.2
    ideal_cov[5:9, 5:9] = 0.6
    for cond in range(5, 9):
        ideal_cov[cond, cond] = 1
    L_full = np.linalg.cholesky(ideal_cov)

    # generating signal
    snr_level = 5.0  # test with high SNR
    inten = np.random.randn(n_V) * 20.0

    # parameters of Gaussian process to generate pseuso SNR
    tau = 0.8
    smooth_width = 5.0
    inten_kernel = 1.0

    coords = np.arange(0, n_V)[:, None]

    dist2 = np.square(coords - coords.T)

    inten_tile = np.tile(inten, [n_V, 1])
    inten_diff2 = (inten_tile - inten_tile.T)**2

    K = np.exp(-dist2 / smooth_width**2 / 2.0 - inten_diff2 /
               inten_kernel**2 / 2.0) * tau**2 + np.eye(n_V) * tau**2 * 0.001

    L = np.linalg.cholesky(K)
    snr = np.exp(np.dot(L, np.random.randn(n_V))) * snr_level
    # Notice that accurately speaking this is not snr. the magnitude of signal
    # depends not only on beta but also on x.
    sqrt_v = noise_level * snr
    betas_simulated = np.dot(L_full, np.random.randn(n_C, n_V)) * sqrt_v
    signal = np.dot(design.design_task, betas_simulated)

    # Adding noise to signal as data
    Y = signal + noise

    scan_onsets = np.linspace(0, design.n_TR, num=n_run + 1)

    # Test fitting with GP prior.
    brsa = BRSA(GP_space=True, GP_inten=True, rank=n_C)

    # Additionally, we test the generation of re-used terms.
    X0 = np.ones(n_T)[:, None]
    D, F, run_TRs, n_run_returned = brsa._prepare_DF(
        n_T, scan_onsets=scan_onsets)
    assert np.shape(D) == (n_T, n_T), 'D has wrong shape'
    assert np.shape(F) == (n_T, n_T), 'F has wrong shape'
    assert np.sum(D) == (n_T - n_run) * 2, 'D is initialized incorrectly.'
    assert np.sum(F) == n_T - n_run * 2, 'F is initialized incorrectly.'
    assert n_run_returned == n_run, (
        'There is mistake in counting number of runs')
    assert np.sum(run_TRs) == n_T, (
        'The segmentation of the total experiment duration is wrong')
    (XTY, XTDY, XTFY, YTY_diag, YTDY_diag, YTFY_diag, XTX, XTDX, XTFX
     ) = brsa._prepare_data_XY(design.design_task, Y, D, F)
    (X0TX0, X0TDX0, X0TFX0, XTX0, XTDX0, XTFX0, X0TY, X0TDY, X0TFY, X0, X_base,
     n_X0, idx_DC
     ) = brsa._prepare_data_XYX0(design.design_task, Y, X0,
                                 np.random.randn(n_T)[:, None], D, F, run_TRs,
                                 no_DC=False)
    assert (np.shape(XTY) == (n_C, n_V)
            and np.shape(XTDY) == (n_C, n_V)
            and np.shape(XTFY) == (n_C, n_V)
            ), 'Dimension of XTY etc. returned from _prepare_data is wrong'
    assert (np.ndim(YTY_diag) == 1
            and np.ndim(YTDY_diag) == 1
            and np.ndim(YTFY_diag) == 1
            ), ("Dimension of YTY_diag etc. returned from _prepare_data is "
                "wrong")
    assert (np.ndim(XTX) == 2
            and np.ndim(XTDX) == 2
            and np.ndim(XTFX) == 2
            ), 'Dimension of XTX etc. returned from _prepare_data is wrong'
    assert (np.ndim(X0TX0) == 2
            and np.ndim(X0TDX0) == 2
            and np.ndim(X0TFX0) == 2
            ), 'Dimension of X0TX0 etc. returned from _prepare_data is wrong'
    assert (np.ndim(XTX0) == 2
            and np.ndim(XTDX0) == 2
            and np.ndim(XTFX0) == 2
            ), 'Dimension of XTX0 etc. returned from _prepare_data is wrong'
    assert (np.ndim(X0TY) == 2
            and np.ndim(X0TDY) == 2
            and np.ndim(X0TFY) == 2
            ), 'Dimension of X0TY etc. returned from _prepare_data is wrong'
    assert (np.shape(X0) == (n_T, n_X0)
            and np.shape(X_base) == (n_T, np.size(idx_DC))
            and np.max(idx_DC) < n_X0
            and np.size(idx_DC) + 1 == n_X0
            ), ("Dimension of X0 or X_base, or n_X0 or indices of DC "
                "components are wrong.")
    l_idx = np.tril_indices(n_C)
    n_l = np.size(l_idx[0])

    # Make sure all the fields are in the indices.
    idx_param_sing, idx_param_fitU, idx_param_fitV = brsa._build_index_param(
        n_l, n_V, 2)
    assert 'Cholesky' in idx_param_sing and 'a1' in idx_param_sing, \
        'The dictionary for parameter indexing misses some keys'
    assert 'Cholesky' in idx_param_fitU and 'a1' in idx_param_fitU, \
        'The dictionary for parameter indexing misses some keys'
    assert 'log_SNR2' in idx_param_fitV and 'c_space' in idx_param_fitV \
        and 'c_inten' in idx_param_fitV and 'c_both' in idx_param_fitV, \
        'The dictionary for parameter indexing misses some keys'

    # Initial parameters are correct parameters with some perturbation
    param0_fitU = np.random.randn(n_l + n_V) * 0.1
    param0_fitV = np.random.randn(n_V + 1) * 0.1
    param0_sing = np.random.randn(n_l + 1) * 0.1
    param0_sing[idx_param_sing['a1']] += np.mean(np.tan(rho1 * np.pi / 2))
    param0_fitV[idx_param_fitV['log_SNR2']] += np.log(snr[:n_V - 1]) * 2
    param0_fitV[idx_param_fitV['c_space']] += np.log(smooth_width) * 2
    param0_fitV[idx_param_fitV['c_inten']] += np.log(inten_kernel) * 2

    # test if the gradients are correct
    # log likelihood and derivative of the _singpara function

    ll0, deriv0 = brsa._loglike_AR1_singpara(param0_sing, XTX, XTDX, XTFX,
                                             YTY_diag, YTDY_diag, YTFY_diag,
                                             XTY, XTDY, XTFY, X0TX0, X0TDX0,
                                             X0TFX0, XTX0, XTDX0, XTFX0, X0TY,
                                             X0TDY, X0TFY, l_idx, n_C, n_T,
                                             n_V, n_run, n_X0, idx_param_sing)
    # We test the gradient to the Cholesky factor
    vec = np.zeros(np.size(param0_sing))
    vec[idx_param_sing['Cholesky'][0]] = 1
    dd = nd.directionaldiff(
        lambda x: brsa._loglike_AR1_singpara(x, XTX, XTDX, XTFX, YTY_diag,
                                             YTDY_diag, YTFY_diag, XTY, XTDY,
                                             XTFY, X0TX0, X0TDX0, X0TFX0, XTX0,
                                             XTDX0, XTFX0, X0TY, X0TDY, X0TFY,
                                             l_idx, n_C, n_T, n_V, n_run, n_X0,
                                             idx_param_sing)[0],
        param0_sing,
        vec)
    assert np.isclose(dd, np.dot(deriv0, vec), rtol=1e-5), (
        'gradient of singpara wrt Cholesky is incorrect')

    # We test the gradient to a1
    vec = np.zeros(np.size(param0_sing))
    vec[idx_param_sing['a1']] = 1
    dd = nd.directionaldiff(
        lambda x: brsa._loglike_AR1_singpara(x, XTX, XTDX, XTFX, YTY_diag,
                                             YTDY_diag, YTFY_diag, XTY, XTDY,
                                             XTFY, X0TX0, X0TDX0, X0TFX0, XTX0,
                                             XTDX0, XTFX0, X0TY, X0TDY, X0TFY,
                                             l_idx, n_C, n_T, n_V, n_run, n_X0,
                                             idx_param_sing)[0],
        param0_sing,
        vec)
    assert np.isclose(dd, np.dot(deriv0, vec),
                      rtol=1e-5), 'gradient of singpara wrt a1 is incorrect'

    # log likelihood and derivative of the fitU function.
    ll0, deriv0 = brsa._loglike_AR1_diagV_fitU(param0_fitU, XTX, XTDX, XTFX,
                                               YTY_diag, YTDY_diag, YTFY_diag,
                                               XTY, XTDY, XTFY, X0TX0, X0TDX0,
                                               X0TFX0, XTX0, XTDX0, XTFX0,
                                               X0TY, X0TDY, X0TFY, np.log(snr)
                                               * 2, l_idx, n_C, n_T, n_V,
                                               n_run, n_X0, idx_param_fitU,
                                               n_C)

    # We test the gradient wrt the reparametrization of AR(1) coefficient of
    # noise.
    vec = np.zeros(np.size(param0_fitU))
    vec[idx_param_fitU['a1'][0]] = 1
    dd = nd.directionaldiff(
        lambda x: brsa._loglike_AR1_diagV_fitU(x, XTX, XTDX, XTFX, YTY_diag,
                                               YTDY_diag, YTFY_diag, XTY, XTDY,
                                               XTFY, X0TX0, X0TDX0, X0TFX0,
                                               XTX0, XTDX0, XTFX0, X0TY, X0TDY,
                                               X0TFY, np.log(snr) * 2, l_idx,
                                               n_C, n_T, n_V, n_run, n_X0,
                                               idx_param_fitU, n_C)[0],
        param0_fitU,
        vec)
    assert np.isclose(dd, np.dot(deriv0, vec), rtol=1e-5), (
        'gradient of fitU wrt to AR(1) coefficient incorrect')

    # We test if the numerical and analytical gradient wrt to the first
    # element of Cholesky factor is correct
    vec = np.zeros(np.size(param0_fitU))
    vec[idx_param_fitU['Cholesky'][0]] = 1
    dd = nd.directionaldiff(
        lambda x: brsa._loglike_AR1_diagV_fitU(x, XTX, XTDX, XTFX, YTY_diag,
                                               YTDY_diag, YTFY_diag, XTY, XTDY,
                                               XTFY, X0TX0, X0TDX0, X0TFX0,
                                               XTX0, XTDX0, XTFX0, X0TY, X0TDY,
                                               X0TFY, np.log(snr) * 2, l_idx,
                                               n_C, n_T, n_V, n_run, n_X0,
                                               idx_param_fitU, n_C)[0],
        param0_fitU,
        vec)
    assert np.isclose(dd, np.dot(deriv0, vec), rtol=1e-5), (
        'gradient of fitU wrt Cholesky factor incorrect')

    # Test on a random direction
    vec = np.random.randn(np.size(param0_fitU))
    vec = vec / np.linalg.norm(vec)
    dd = nd.directionaldiff(
        lambda x: brsa._loglike_AR1_diagV_fitU(x, XTX, XTDX, XTFX, YTY_diag,
                                               YTDY_diag, YTFY_diag, XTY, XTDY,
                                               XTFY, X0TX0, X0TDX0, X0TFX0,
                                               XTX0, XTDX0, XTFX0, X0TY, X0TDY,
                                               X0TFY, np.log(snr) * 2, l_idx,
                                               n_C, n_T, n_V, n_run, n_X0,
                                               idx_param_fitU, n_C)[0],
        param0_fitU,
        vec)
    assert np.isclose(dd, np.dot(deriv0, vec),
                      rtol=1e-5), 'gradient of fitU incorrect'

    # We test the gradient of _fitV wrt to log(SNR^2) assuming no GP prior.
    X0TAX0, XTAX0, X0TAY, X0TAX0_i, \
        XTAcorrX, XTAcorrY, YTAcorrY, LTXTAcorrY, XTAcorrXL, LTXTAcorrXL = \
        brsa._precompute_ar1_quad_forms(XTY, XTDY, XTFY,
                                        YTY_diag, YTDY_diag, YTFY_diag,
                                        XTX, XTDX, XTFX,
                                        X0TX0, X0TDX0, X0TFX0,
                                        XTX0, XTDX0, XTFX0,
                                        X0TY, X0TDY, X0TFY,
                                        L_full, rho1, n_V, n_X0)
    assert np.shape(XTAcorrX) == (n_V, n_C, n_C), (
        'Dimension of XTAcorrX is wrong by _precompute_ar1_quad_forms()')
    assert XTAcorrY.shape == XTY.shape, (
        'Shape of XTAcorrY is wrong by _precompute_ar1_quad_forms()')
    assert YTAcorrY.shape == YTY_diag.shape, (
        'Shape of YTAcorrY is wrong by _precompute_ar1_quad_forms()')
    assert np.shape(X0TAX0) == (n_V, n_X0, n_X0), (
        'Dimension of X0TAX0 is wrong by _precompute_ar1_quad_forms()')
    assert np.shape(XTAX0) == (n_V, n_C, n_X0), (
        'Dimension of XTAX0 is wrong by _precompute_ar1_quad_forms()')
    assert X0TAY.shape == X0TY.shape, (
        'Shape of X0TAX0 is wrong by _precompute_ar1_quad_forms()')
    assert np.all(np.isfinite(X0TAX0_i)), (
        'Inverse of X0TAX0 includes NaN or Inf')
    ll0, deriv0 = brsa._loglike_AR1_diagV_fitV(
        param0_fitV[idx_param_fitV['log_SNR2']], X0TAX0, XTAX0, X0TAY,
        X0TAX0_i, XTAcorrX, XTAcorrY, YTAcorrY, LTXTAcorrY, XTAcorrXL,
        LTXTAcorrXL, L_full[l_idx], np.tan(rho1 * np.pi / 2), l_idx, n_C, n_T,
        n_V, n_run, n_X0, idx_param_fitV, n_C, False, False)
    vec = np.zeros(np.size(param0_fitV[idx_param_fitV['log_SNR2']]))
    vec[idx_param_fitV['log_SNR2'][0]] = 1
    dd = nd.directionaldiff(
        lambda x: brsa._loglike_AR1_diagV_fitV(x, X0TAX0, XTAX0, X0TAY,
                                               X0TAX0_i, XTAcorrX, XTAcorrY,
                                               YTAcorrY, LTXTAcorrY, XTAcorrXL,
                                               LTXTAcorrXL, L_full[l_idx],
                                               np.tan(rho1 * np.pi / 2),
                                               l_idx, n_C, n_T, n_V, n_run,
                                               n_X0, idx_param_fitV, n_C,
                                               False, False)[0],
        param0_fitV[idx_param_fitV['log_SNR2']],
        vec)
    assert np.isclose(dd, np.dot(deriv0, vec), rtol=1e-5), (
        'gradient of fitV wrt log(SNR2) incorrect for model without GP')

    # We test the gradient of _fitV wrt to log(SNR^2) assuming GP prior.
    ll0, deriv0 = brsa._loglike_AR1_diagV_fitV(
        param0_fitV, X0TAX0, XTAX0, X0TAY, X0TAX0_i, XTAcorrX, XTAcorrY,
        YTAcorrY, LTXTAcorrY, XTAcorrXL, LTXTAcorrXL, L_full[l_idx],
        np.tan(rho1 * np.pi / 2), l_idx, n_C, n_T, n_V, n_run, n_X0,
        idx_param_fitV, n_C, True, True, dist2, inten_diff2, 100, 100)
    vec = np.zeros(np.size(param0_fitV))
    vec[idx_param_fitV['log_SNR2'][0]] = 1
    dd = nd.directionaldiff(
        lambda x: brsa._loglike_AR1_diagV_fitV(x, X0TAX0, XTAX0, X0TAY,
                                               X0TAX0_i, XTAcorrX, XTAcorrY,
                                               YTAcorrY, LTXTAcorrY, XTAcorrXL,
                                               LTXTAcorrXL, L_full[l_idx],
                                               np.tan(rho1 * np.pi / 2),
                                               l_idx, n_C, n_T, n_V, n_run,
                                               n_X0, idx_param_fitV, n_C, True,
                                               True, dist2, inten_diff2, 100,
                                               100)[0],
        param0_fitV,
        vec)
    assert np.isclose(dd, np.dot(deriv0, vec), rtol=1e-5), (
        'gradient of fitV srt log(SNR2) incorrect for model with GP')

    # We test the graident wrt spatial length scale parameter of GP prior
    vec = np.zeros(np.size(param0_fitV))
    vec[idx_param_fitV['c_space']] = 1
    dd = nd.directionaldiff(
        lambda x: brsa._loglike_AR1_diagV_fitV(x, X0TAX0, XTAX0, X0TAY,
                                               X0TAX0_i, XTAcorrX, XTAcorrY,
                                               YTAcorrY, LTXTAcorrY, XTAcorrXL,
                                               LTXTAcorrXL, L_full[l_idx],
                                               np.tan(rho1 * np.pi / 2),
                                               l_idx, n_C, n_T, n_V, n_run,
                                               n_X0, idx_param_fitV, n_C, True,
                                               True, dist2, inten_diff2, 100,
                                               100)[0],
        param0_fitV,
        vec)
    assert np.isclose(dd, np.dot(deriv0, vec), rtol=1e-5), (
        'gradient of fitV wrt spatial length scale of GP incorrect')

    # We test the graident wrt intensity length scale parameter of GP prior
    vec = np.zeros(np.size(param0_fitV))
    vec[idx_param_fitV['c_inten']] = 1
    dd = nd.directionaldiff(
        lambda x: brsa._loglike_AR1_diagV_fitV(x, X0TAX0, XTAX0, X0TAY,
                                               X0TAX0_i, XTAcorrX, XTAcorrY,
                                               YTAcorrY, LTXTAcorrY, XTAcorrXL,
                                               LTXTAcorrXL, L_full[l_idx],
                                               np.tan(rho1 * np.pi / 2),
                                               l_idx, n_C, n_T, n_V, n_run,
                                               n_X0, idx_param_fitV, n_C, True,
                                               True, dist2, inten_diff2, 100,
                                               100)[0],
        param0_fitV,
        vec)
    assert np.isclose(dd, np.dot(deriv0, vec), rtol=1e-5), (
        'gradient of fitV wrt intensity length scale of GP incorrect')

    # We test the graident on a random direction
    vec = np.random.randn(np.size(param0_fitV))
    vec = vec / np.linalg.norm(vec)
    dd = nd.directionaldiff(
        lambda x: brsa._loglike_AR1_diagV_fitV(x, X0TAX0, XTAX0, X0TAY,
                                               X0TAX0_i, XTAcorrX, XTAcorrY,
                                               YTAcorrY, LTXTAcorrY, XTAcorrXL,
                                               LTXTAcorrXL, L_full[l_idx],
                                               np.tan(rho1 * np.pi / 2),
                                               l_idx, n_C, n_T, n_V, n_run,
                                               n_X0, idx_param_fitV, n_C, True,
                                               True, dist2, inten_diff2, 100,
                                               100)[0],
        param0_fitV,
        vec)
    assert np.isclose(dd, np.dot(deriv0, vec), rtol=1e-5), (
        'gradient of fitV incorrect')


def test_nureg_determine():
    from brainiak.reprsimil.brsa import Ncomp_SVHT_MG_DLD_approx
    import numpy as np
    x = np.dot(np.random.randn(100, 5), np.random.randn(5, 40)) + \
        np.random.randn(100, 40) * 0.01
    ncomp = Ncomp_SVHT_MG_DLD_approx(x)
    assert ncomp >= 3 and ncomp <= 8, (
        'recovered number of components should be in a reasonable range')


def test_half_log_det():
    import numpy as np
    from brainiak.reprsimil.brsa import BRSA
    a = np.asarray([[1, 0.2], [0.2, 1]])
    brsa = BRSA()
    half_log_det = np.log(np.linalg.det(a)) / 2
    assert np.isclose(half_log_det, brsa._half_log_det(
        a)), 'half log determinant function is wrong'


def test_n_nureg():
    import brainiak.reprsimil.brsa
    import numpy as np
    # noise = np.random.randn(100,30)
    noise = np.dot(np.random.randn(100, 8), np.random.randn(
        8, 30)) + np.random.randn(100, 30) * 0.01
    design = np.random.randn(100, 2)
    s = brainiak.reprsimil.brsa.BRSA(n_iter=2)
    s.fit(X=noise, design=design)
    assert s.n_nureg_ > 2 and s.n_nureg_ < 16, (
        'n_nureg_ estimation is wrong in BRSA')



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pytest

from brainiak.fcma.mvpa_voxelselector import MVPAVoxelSelector
from brainiak.searchlight.searchlight import Searchlight
from sklearn import svm
import numpy as np
from mpi4py import MPI
from numpy.random import RandomState


# specify the random state to fix the random numbers
prng = RandomState(1234567890)


@pytest.mark.mpiexec(n=2)
def test_mvpa_voxel_selection(pool_size):
    data = prng.rand(5, 5, 5, 8).astype(np.float32)
    # all MPI processes read the mask; the mask file is small
    mask = np.ones([5, 5, 5], dtype=bool)
    mask[0, 0, :] = False
    labels = [0, 1, 0, 1, 0, 1, 0, 1]
    # 2 subjects, 4 epochs per subject
    sl = Searchlight(sl_rad=1, pool_size=pool_size)
    mvs = MVPAVoxelSelector(data, mask, labels, 2, sl)
    # for cross validation, use SVM with precomputed kernel

    clf = svm.SVC(kernel='rbf', C=10, gamma='auto')
    result_volume, results = mvs.run(clf)
    if MPI.COMM_WORLD.Get_rank() == 0:
        output = []
        for tuple in results:
            if tuple[1] > 0:
                output.append(int(8*tuple[1]))
        expected_output = [6, 6, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4,
                           4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 1]
        assert np.allclose(output, expected_output, atol=1), \
            'voxel selection via SVM does not provide correct results'


if __name__ == '__main__':
    test_mvpa_voxel_selection()



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pytest

from brainiak.fcma.voxelselector import VoxelSelector
from scipy.stats.mstats import zscore
from sklearn import svm
from sklearn.linear_model import LogisticRegression
import numpy as np
import math
from mpi4py import MPI
from numpy.random import RandomState


def create_epoch(prng):
    row = 12
    col = 5
    mat = prng.rand(row, col).astype(np.float32)
    mat = zscore(mat, axis=0, ddof=0)
    # if zscore fails (standard deviation is zero),
    # set all values to be zero
    mat = np.nan_to_num(mat)
    mat = mat / math.sqrt(mat.shape[0])
    return mat


@pytest.mark.mpiexec(n=2)
def test_voxel_selection(pool_size):

    # For VoxelSelector, process_num=0 means no multiprocessing
    if pool_size == 1:
        process_num = 0
    else:
        process_num = pool_size

    prng = RandomState(1234567890)
    fake_raw_data = [create_epoch(prng) for i in range(8)]
    labels = [0, 1, 0, 1, 0, 1, 0, 1]
    # 2 subjects, 4 epochs per subject
    vs = VoxelSelector(labels, 4, 2, fake_raw_data, voxel_unit=1,
                       process_num=process_num)
    # test scipy normalization
    fake_corr = prng.rand(1, 4, 5).astype(np.float32)
    fake_corr = vs._correlation_normalization(fake_corr)
    if MPI.COMM_WORLD.Get_rank() == 0:
        expected_fake_corr = [[[1.06988919, 0.51641309, -0.46790636,
                                -1.31926763, 0.2270218],
                               [-1.22142744, -1.39881694, -1.2979387,
                                1.05702305, -0.6525566],
                               [0.89795232, 1.27406132, 0.36460185,
                                0.87538344, 1.5227468],
                               [-0.74641371, -0.39165771, 1.40124381,
                                -0.61313909, -1.0972116]]]
        assert np.allclose(fake_corr, expected_fake_corr), \
            'within-subject normalization does not provide correct results'
    # for cross validation, use SVM with precomputed kernel
    # no shrinking, set C=1
    clf = svm.SVC(kernel='precomputed', shrinking=False, C=1, gamma='auto')
    results = vs.run(clf)
    if MPI.COMM_WORLD.Get_rank() == 0:
        output = [None] * len(results)
        for tuple in results:
            output[tuple[0]] = int(8*tuple[1])
        expected_output = [7, 4, 6, 4, 4]
        assert np.allclose(output, expected_output, atol=1), \
            'voxel selection via SVM does not provide correct results'
    # for cross validation, use logistic regression
    clf = LogisticRegression()
    results = vs.run(clf)
    if MPI.COMM_WORLD.Get_rank() == 0:
        output = [None] * len(results)
        for tuple in results:
            output[tuple[0]] = int(8*tuple[1])
        expected_output = [6, 3, 6, 4, 4]
        assert np.allclose(output, expected_output, atol=1), (
            "voxel selection via logistic regression does not provide correct "
            "results")


@pytest.mark.mpiexec(n=2)
def test_voxel_selection_with_two_masks(pool_size):

    # For VoxelSelector, process_num=0 means no multiprocessing
    if pool_size == 1:
        process_num = 0
    else:
        process_num = pool_size

    prng = RandomState(1234567890)
    fake_raw_data1 = [create_epoch(prng) for i in range(8)]
    fake_raw_data2 = [create_epoch(prng) for i in range(8)]
    labels = [0, 1, 0, 1, 0, 1, 0, 1]
    # 2 subjects, 4 epochs per subject
    vs = VoxelSelector(labels, 4, 2, fake_raw_data1,
                       raw_data2=fake_raw_data2, voxel_unit=1,
                       process_num=process_num)
    # for cross validation, use SVM with precomputed kernel
    # no shrinking, set C=1
    clf = svm.SVC(kernel='precomputed', shrinking=False, C=1, gamma='auto')
    results = vs.run(clf)
    if MPI.COMM_WORLD.Get_rank() == 0:
        output = [None] * len(results)
        for tuple in results:
            output[tuple[0]] = int(8*tuple[1])
        expected_output = [3, 3, 7, 5, 7]
        assert np.allclose(output, expected_output, atol=1), \
            'voxel selection via SVM does not provide correct results'
    # for cross validation, use logistic regression
    clf = LogisticRegression()
    results = vs.run(clf)
    if MPI.COMM_WORLD.Get_rank() == 0:
        output = [None] * len(results)
        for tuple in results:
            output[tuple[0]] = int(8*tuple[1])
        expected_output = [4, 3, 7, 4, 6]
        assert np.allclose(output, expected_output, atol=1), (
            "voxel selection via logistic regression does not provide correct "
            "results")


def test_dummy():
    """
    This is a dummy test to work around for the issue of pytest and
    pytest-mpiexec. See here the discussion of the same issue in
    pytest-forked:

    https://github.com/pytest-dev/pytest-forked/issues/67#issuecomment-1964718720
    """
    pass


if __name__ == '__main__':
    test_voxel_selection()
    test_voxel_selection_with_two_masks()



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import numpy as np
from numpy.random import RandomState
from brainiak.fcma.util import compute_correlation

# specify the random state to fix the random numbers
prng = RandomState(1234567890)


def test_correlation_computation():
    row1 = 5
    col = 10
    row2 = 6
    mat1 = prng.rand(row1, col).astype(np.float32)
    mat2 = prng.rand(row2, col).astype(np.float32)
    corr = compute_correlation(mat1, mat1)
    expected_corr = np.corrcoef(mat1)
    assert np.allclose(corr, expected_corr, atol=1e-5), (
        "high performance correlation computation does not provide correct "
        "correlation results within the same set")
    corr = compute_correlation(mat1, mat2)
    mat = np.concatenate((mat1, mat2), axis=0)
    expected_corr = np.corrcoef(mat)[0:row1, row1:]
    assert np.allclose(corr, expected_corr, atol=1e-5), (
        "high performance correlation computation does not provide correct "
        "correlation results between two sets")


def test_correlation_nans():
    row1 = 5
    col = 10
    row2 = 6
    mat1 = prng.rand(row1, col).astype(np.float32)
    mat2 = prng.rand(row2, col).astype(np.float32)
    mat1[0, 0] = np.nan
    corr = compute_correlation(mat1, mat2, return_nans=False)
    assert np.all(corr == 0, axis=1)[0]
    assert np.sum(corr == 0) == row2
    corr = compute_correlation(mat1, mat2, return_nans=True)
    assert np.all(np.isnan(corr), axis=1)[0]
    assert np.sum(np.isnan(corr)) == row2


if __name__ == '__main__':
    test_correlation_computation()
    test_correlation_nans()



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from sklearn.exceptions import NotFittedError
import pytest


def test_can_instantiate(tmp_path):
    import brainiak.funcalign.srm
    s = brainiak.funcalign.srm.SRM()
    assert s, "Invalid SRM instance!"

    import numpy as np
    np.random.seed(0)

    voxels = 100
    samples = 500
    subjects = 2
    features = 3

    s = brainiak.funcalign.srm.SRM(n_iter=5, features=features)
    assert s, "Invalid SRM instance!"

    # Create a Shared response S with K = 3
    theta = np.linspace(-4 * np.pi, 4 * np.pi, samples)
    z = np.linspace(-2, 2, samples)
    r = z**2 + 1
    x = r * np.sin(theta)
    y = r * np.cos(theta)

    S = np.vstack((x, y, z))

    X = []
    W = []
    Q, R = np.linalg.qr(np.random.random((voxels, features)))
    W.append(Q)
    X.append(Q.dot(S) + 0.1*np.random.random((voxels, samples)))

    # Check that transform does NOT run before fitting the model
    with pytest.raises(NotFittedError):
        s.transform(X)
    print("Test: transforming before fitting the model")

    # Check that it does NOT run with 1 subject
    with pytest.raises(ValueError):
        s.fit(X)
    print("Test: running SRM with 1 subject")

    for subject in range(1, subjects):
        Q, R = np.linalg.qr(np.random.random((voxels, features)))
        W.append(Q)
        X.append(Q.dot(S) + 0.1*np.random.random((voxels, samples)))

    # Check that runs with 2 subject
    s.fit(X)
    from pathlib import Path
    sr_v0_4 = np.load(Path(__file__).parent / "sr_v0_4.npz")['sr']
    assert (np.allclose(sr_v0_4, s.s_))

    assert len(s.w_) == subjects, (
        "Invalid computation of SRM! (wrong # subjects in W)")
    for subject in range(subjects):
        assert s.w_[subject].shape[0] == voxels, (
            "Invalid computation of SRM! (wrong # voxels in W)")
        assert s.w_[subject].shape[1] == features, (
            "Invalid computation of SRM! (wrong # features in W)")
        ortho = np.linalg.norm(s.w_[subject].T.dot(s.w_[subject])
                               - np.eye(s.w_[subject].shape[1]),
                               'fro')
        assert ortho < 1e-7, "A Wi mapping is not orthonormal in SRM."
        difference = np.linalg.norm(X[subject] - s.w_[subject].dot(s.s_),
                                    'fro')
        datanorm = np.linalg.norm(X[subject], 'fro')
        assert difference/datanorm < 1.0, "Model seems incorrectly computed."
    assert s.s_.shape[0] == features, (
        "Invalid computation of SRM! (wrong # features in S)")
    assert s.s_.shape[1] == samples, (
        "Invalid computation of SRM! (wrong # samples in S)")

    # Check that it does run to compute the shared response after the model
    # computation
    new_s = s.transform(X)

    assert len(new_s) == subjects, (
        "Invalid computation of SRM! (wrong # subjects after transform)")
    for subject in range(subjects):
        assert new_s[subject].shape[0] == features, (
            "Invalid computation of SRM! (wrong # features after transform)")
        assert new_s[subject].shape[1] == samples, (
            "Invalid computation of SRM! (wrong # samples after transform)")

    # Check that it does NOT run with non-matching number of subjects
    with pytest.raises(ValueError):
        s.transform(X[1])
    print("Test: transforming with non-matching number of subjects")

    # Check that it does not run without enough samples (TRs).
    with pytest.raises(ValueError):
        s.set_params(features=(samples+1))
        s.fit(X)
    print("Test: not enough samples")

    # Check that it does not run with different number of samples (TRs)
    S2 = S[:, :-2]
    X.append(Q.dot(S2))
    with pytest.raises(ValueError):
        s.fit(X)
    print("Test: different number of samples per subject")

    # Check save/load functionality for fitted SRM
    srm_path = tmp_path / 'srm.npz'
    s.save(srm_path)
    s_load = brainiak.funcalign.srm.load(srm_path)
    assert np.array_equal(s.s_, s_load.s_)
    for w, wl in zip(s.w_, s_load.w_):
        assert np.array_equal(w, wl)
    assert np.array_equal(s.sigma_s_, s_load.sigma_s_)
    assert np.array_equal(s.mu_, s_load.mu_)
    assert np.array_equal(s.rho2_, s_load.rho2_)
    assert s.features == s_load.features
    assert s.n_iter == s_load.n_iter
    assert s.rand_seed == s_load.rand_seed
    print("Test: save/load functionality")


def test_new_subject():
    import brainiak.funcalign.srm
    s = brainiak.funcalign.srm.SRM()
    assert s, "Invalid SRM instance!"

    import numpy as np
    np.random.seed(0)

    voxels = 100
    samples = 500
    subjects = 3
    features = 3

    s = brainiak.funcalign.srm.SRM(n_iter=5, features=features)
    assert s, "Invalid SRM instance!"

    # Create a Shared response S with K = 3
    theta = np.linspace(-4 * np.pi, 4 * np.pi, samples)
    z = np.linspace(-2, 2, samples)
    r = z**2 + 1
    x = r * np.sin(theta)
    y = r * np.cos(theta)

    S = np.vstack((x, y, z))

    X = []
    W = []
    Q, R = np.linalg.qr(np.random.random((voxels, features)))
    W.append(Q)
    X.append(Q.dot(S) + 0.1*np.random.random((voxels, samples)))

    for subject in range(1, subjects):
        Q, R = np.linalg.qr(np.random.random((voxels, features)))
        W.append(Q)
        X.append(Q.dot(S) + 0.1*np.random.random((voxels, samples)))

    # Check that transform does NOT run before fitting the model
    with pytest.raises(NotFittedError):
        s.transform_subject(X)
    print("Test: transforming before fitting the model")

    # Check that runs with 3 subject
    s.fit(X)

    # Check that you get an error when the data is the wrong shape
    with pytest.raises(ValueError):
        s.transform_subject(X[0].T)

    # Check that it does run to compute a new subject
    new_w = s.transform_subject(X[0])
    assert new_w.shape[1] == features, (
            "Invalid computation of SRM! (wrong # features for new subject)")
    assert new_w.shape[0] == voxels, (
            "Invalid computation of SRM! (wrong # voxels for new subject)")

    # Check that these analyses work with the deterministic SRM too
    ds = brainiak.funcalign.srm.DetSRM(n_iter=5, features=features)

    # Check that transform does NOT run before fitting the model
    with pytest.raises(NotFittedError):
        ds.transform_subject(X)
    print("Test: transforming before fitting the model")

    # Check that runs with 3 subject
    ds.fit(X)

    # Check that you get an error when the data is the wrong shape
    with pytest.raises(ValueError):
        ds.transform_subject(X[0].T)

    # Check that it does run to compute a new subject
    new_w = ds.transform_subject(X[0])
    assert new_w.shape[1] == features, (
            "Invalid computation of SRM! (wrong # features for new subject)")
    assert new_w.shape[0] == voxels, (
            "Invalid computation of SRM! (wrong # voxels for new subject)")


def test_det_srm():
    import brainiak.funcalign.srm
    model = brainiak.funcalign.srm.DetSRM()
    assert model, "Invalid DetSRM instance!"

    import numpy as np

    voxels = 100
    samples = 500
    subjects = 2
    features = 3

    model = brainiak.funcalign.srm.DetSRM(n_iter=5, features=features)
    assert model, "Invalid DetSRM instance!"

    # Create a Shared response S with K = 3
    theta = np.linspace(-4 * np.pi, 4 * np.pi, samples)
    z = np.linspace(-2, 2, samples)
    r = z**2 + 1
    x = r * np.sin(theta)
    y = r * np.cos(theta)

    S = np.vstack((x, y, z))

    X = []
    W = []
    Q, R = np.linalg.qr(np.random.random((voxels, features)))
    W.append(Q)
    X.append(Q.dot(S) + 0.1*np.random.random((voxels, samples)))

    # Check that transform does NOT run before fitting the model
    with pytest.raises(NotFittedError):
        model.transform(X)
    print("Test: transforming before fitting the model")

    # Check that it does NOT run with 1 subject
    with pytest.raises(ValueError):
        model.fit(X)
    print("Test: running DetSRM with 1 subject")

    for subject in range(1, subjects):
        Q, R = np.linalg.qr(np.random.random((voxels, features)))
        W.append(Q)
        X.append(Q.dot(S) + 0.1*np.random.random((voxels, samples)))

    # Check that runs with 2 subject
    model.fit(X)

    assert len(model.w_) == subjects, (
        "Invalid computation of DetSRM! (wrong # subjects in W)")
    for subject in range(subjects):
        assert model.w_[subject].shape[0] == voxels, (
            "Invalid computation of DetSRM! (wrong # voxels in W)")
        assert model.w_[subject].shape[1] == features, (
            "Invalid computation of DetSRM! (wrong # features in W)")
        ortho = np.linalg.norm(model.w_[subject].T.dot(model.w_[subject])
                               - np.eye(model.w_[subject].shape[1]),
                               'fro')
        assert ortho < 1e-7, "A Wi mapping is not orthonormal in DetSRM."
        difference = np.linalg.norm(X[subject]
                                    - model.w_[subject].dot(model.s_),
                                    'fro')
        datanorm = np.linalg.norm(X[subject], 'fro')
        assert difference/datanorm < 1.0, "Model seems incorrectly computed."
    assert model.s_.shape[0] == features, (
        "Invalid computation of DetSRM! (wrong # features in S)")
    assert model.s_.shape[1] == samples, (
        "Invalid computation of DetSRM! (wrong # samples in S)")

    # Check that it does run to compute the shared response after the model
    # computation
    new_s = model.transform(X)

    assert len(new_s) == subjects, (
        "Invalid computation of DetSRM! (wrong # subjects after transform)")
    for subject in range(subjects):
        assert new_s[subject].shape[0] == features, (
            "Invalid computation of DetSRM! (wrong # features after "
            "transform)")
        assert new_s[subject].shape[1] == samples, (
            "Invalid computation of DetSRM! (wrong # samples after transform)")

    # Check that it does run to compute a new subject
    new_w = model.transform_subject(X[0])
    assert new_w.shape[1] == features, (
            "Invalid computation of SRM! (wrong # features for new subject)")
    assert new_w.shape[0] == voxels, (
            "Invalid computation of SRM! (wrong # voxels for new subject)")

    # Check that it does NOT run with non-matching number of subjects
    with pytest.raises(ValueError):
        model.transform(X[1])
    print("Test: transforming with non-matching number of subjects")

    # Check that it does not run without enough samples (TRs).
    with pytest.raises(ValueError):
        model.set_params(features=(samples+1))
        model.fit(X)
    print("Test: not enough samples")

    # Check that it does not run with different number of samples (TRs)
    S2 = S[:, :-2]
    X.append(Q.dot(S2))
    with pytest.raises(ValueError):
        model.fit(X)
    print("Test: different number of samples per subject")



import pytest
import numpy as np
from scipy.stats import norm, wishart, pearsonr

from brainiak.matnormal.covs import (
    CovIdentity,
    CovUnconstrainedCholesky,
    CovUnconstrainedInvCholesky,
    CovDiagonal,
)
from brainiak.matnormal.regression import MatnormalRegression
from brainiak.matnormal.utils import rmn

m = 100
n = 4
p = 5

corrtol = 0.8  # at least this much correlation between true and est to pass


def test_matnorm_regression_unconstrained(seeded_rng):

    # Y = XB + eps
    # Y is m x p, B is n x p, eps is m x p
    X = norm.rvs(size=(m, n))
    B = norm.rvs(size=(n, p))
    Y_hat = X.dot(B)
    rowcov_true = np.eye(m)
    colcov_true = wishart.rvs(p + 2, np.eye(p))

    Y = Y_hat + rmn(rowcov_true, colcov_true)

    row_cov = CovIdentity(size=m)
    col_cov = CovUnconstrainedCholesky(size=p)

    model = MatnormalRegression(time_cov=row_cov, space_cov=col_cov)

    model.fit(X, Y, naive_init=False)

    assert pearsonr(B.flatten(), model.beta_.flatten())[0] >= corrtol

    pred_y = model.predict(X)
    assert pearsonr(pred_y.flatten(), Y_hat.flatten())[0] >= corrtol

    model = MatnormalRegression(time_cov=row_cov, space_cov=col_cov)

    model.fit(X, Y, naive_init=True)

    assert pearsonr(B.flatten(), model.beta_.flatten())[0] >= corrtol

    pred_y = model.predict(X)
    assert pearsonr(pred_y.flatten(), Y_hat.flatten())[0] >= corrtol


def test_matnorm_regression_unconstrainedprec(seeded_rng):

    # Y = XB + eps
    # Y is m x n, B is n x p, eps is m x p
    X = norm.rvs(size=(m, n))
    B = norm.rvs(size=(n, p))
    Y_hat = X.dot(B)
    rowcov_true = np.eye(m)
    colcov_true = wishart.rvs(p + 2, np.eye(p))

    Y = Y_hat + rmn(rowcov_true, colcov_true)

    row_cov = CovIdentity(size=m)
    col_cov = CovUnconstrainedInvCholesky(size=p)

    model = MatnormalRegression(time_cov=row_cov, space_cov=col_cov)

    model.fit(X, Y, naive_init=False)

    assert pearsonr(B.flatten(), model.beta_.flatten())[0] >= corrtol

    pred_y = model.predict(X)
    assert pearsonr(pred_y.flatten(), Y_hat.flatten())[0] >= corrtol


def test_matnorm_regression_optimizerChoice(seeded_rng):

    # Y = XB + eps
    # Y is m x n, B is n x p, eps is m x p
    X = norm.rvs(size=(m, n))
    B = norm.rvs(size=(n, p))
    Y_hat = X.dot(B)
    rowcov_true = np.eye(m)
    colcov_true = wishart.rvs(p + 2, np.eye(p))

    Y = Y_hat + rmn(rowcov_true, colcov_true)

    row_cov = CovIdentity(size=m)
    col_cov = CovUnconstrainedInvCholesky(size=p)

    model = MatnormalRegression(time_cov=row_cov, space_cov=col_cov,
                                optimizer="CG")

    model.fit(X, Y, naive_init=False)

    assert pearsonr(B.flatten(), model.beta_.flatten())[0] >= corrtol

    pred_y = model.predict(X)
    assert pearsonr(pred_y.flatten(), Y_hat.flatten())[0] >= corrtol


def test_matnorm_regression_scaledDiag(seeded_rng):

    # Y = XB + eps
    # Y is m x n, B is n x p, eps is m x p
    X = norm.rvs(size=(m, n))
    B = norm.rvs(size=(n, p))
    Y_hat = X.dot(B)

    rowcov_true = np.eye(m)
    colcov_true = np.diag(np.abs(norm.rvs(size=p)))

    Y = Y_hat + rmn(rowcov_true, colcov_true)

    row_cov = CovIdentity(size=m)
    col_cov = CovDiagonal(size=p)

    model = MatnormalRegression(time_cov=row_cov, space_cov=col_cov)

    model.fit(X, Y, naive_init=False)

    assert pearsonr(B.flatten(), model.beta_.flatten())[0] >= corrtol

    pred_y = model.predict(X)
    assert pearsonr(pred_y.flatten(), Y_hat.flatten())[0] >= corrtol

    # we only do calibration test on the scaled diag
    # model because to hit corrtol on unconstrainedCov
    # we'd need a lot more data, which would make the test slow
    X_hat = model.calibrate(Y)
    assert pearsonr(X_hat.flatten(), X.flatten())[0] >= corrtol


def test_matnorm_calibration_raises(seeded_rng):

    # Y = XB + eps
    # Y is m x n, B is n x p, eps is m x p
    X = norm.rvs(size=(2, 5))
    B = norm.rvs(size=(5, 3))
    Y_hat = X.dot(B)

    rowcov_true = np.eye(2)
    colcov_true = np.diag(np.abs(norm.rvs(size=3)))

    Y = Y_hat + rmn(rowcov_true, colcov_true)

    row_cov = CovIdentity(size=2)
    col_cov = CovDiagonal(size=3)

    model = MatnormalRegression(time_cov=row_cov, space_cov=col_cov)

    model.fit(X, Y, naive_init=False)

    with pytest.raises(RuntimeError):
        model.calibrate(Y)



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import numpy as np
from mpi4py import MPI
import sys

from brainiak.searchlight.searchlight import Searchlight
from brainiak.searchlight.searchlight import Diamond

"""Distributed Searchlight Example
example usage: mpirun -n 4 python3 example_searchlight.py
"""

comm = MPI.COMM_WORLD
rank = comm.rank
size = comm.size

# Dataset size parameters
dim = 40
ntr = 400
maskrad = 15

# Predictive point parameters
pt = (23,23,23)
kernel_dim = 5
weight = 1

# Generate data
data = np.random.random((dim,dim,dim,ntr)) if rank == 0 else None
mask = np.zeros((dim,dim,dim), dtype=bool)
for i in range(dim):
  for j in range(dim):
    for k in range(dim):
      dist = np.sqrt(((dim/2)-i)**2 + ((dim/2)-j)**2 + ((dim/2)-k)**2)
      if(dist < maskrad):
        mask[i,j,k] = 1

# Generate labels
labels = np.random.choice([True, False], (ntr,)) if rank == 0 else None

# Inject predictive region in random data
if rank == 0:
  kernel = np.zeros((kernel_dim,kernel_dim,kernel_dim))
  for i in range(kernel_dim):
    for j in range(kernel_dim):
      for k in range(kernel_dim):
        arr = np.array([i-(kernel_dim/2),j-(kernel_dim/2),k-(kernel_dim/2)])
        kernel [i,j,k] = np.exp(-np.dot(arr.T,arr))
  kernel = kernel / np.sum(kernel)

  for (idx, l) in enumerate(labels):
    if l:
      data[pt[0]:pt[0]+kernel_dim,pt[1]:pt[1]+kernel_dim,pt[2]:pt[2]+kernel_dim,idx] += kernel * weight
    else:
      data[pt[0]:pt[0]+kernel_dim,pt[1]:pt[1]+kernel_dim,pt[2]:pt[2]+kernel_dim,idx] -= kernel * weight

# Create searchlight object
sl = Searchlight(sl_rad=1, max_blk_edge=5, shape=Diamond,
                 min_active_voxels_proportion=0)

# Distribute data to processes
sl.distribute([data], mask)
sl.broadcast(labels)

# Define voxel function
def sfn(l, msk, myrad, bcast_var):
  import sklearn.svm
  import sklearn.model_selection
  classifier = sklearn.svm.SVC(gamma='auto')
  data = l[0][msk,:].T
  return np.mean(sklearn.model_selection.cross_val_score(classifier, data, bcast_var,n_jobs=1))

# Run searchlight
global_outputs = sl.run_searchlight(sfn)

# Visualize result
if rank == 0:
  print(global_outputs)
  global_outputs = np.array(global_outputs, dtype=np.float)
  import matplotlib.pyplot as plt
  for (cnt, img) in enumerate(global_outputs):
    plt.imshow(img,cmap='hot',vmin=0,vmax=1)
    plt.savefig('img' + str(cnt) + '.png')
    plt.clf()








import numpy as np
from numpy.testing import assert_allclose
from scipy.stats import multivariate_normal
import tensorflow as tf

from brainiak.matnormal.utils import rmn
from brainiak.matnormal.matnormal_likelihoods import matnorm_logp
from brainiak.matnormal.covs import CovIdentity, CovUnconstrainedCholesky

# X is m x n, so A sould be m x p

m = 5
n = 4
p = 3

rtol = 1e-7


def test_against_scipy_mvn_row(seeded_rng):

    rowcov = CovUnconstrainedCholesky(size=m)
    colcov = CovIdentity(size=n)
    X = rmn(np.eye(m), np.eye(n))
    X_tf = tf.constant(X, "float64")

    rowcov_np = rowcov._cov

    scipy_answer = np.sum(multivariate_normal.logpdf(
        X.T, np.zeros([m]), rowcov_np))
    tf_answer = matnorm_logp(X_tf, rowcov, colcov)
    assert_allclose(scipy_answer, tf_answer, rtol=rtol)


def test_against_scipy_mvn_col(seeded_rng):

    rowcov = CovIdentity(size=m)
    colcov = CovUnconstrainedCholesky(size=n)
    X = rmn(np.eye(m), np.eye(n))
    X_tf = tf.constant(X, "float64")

    colcov_np = colcov._cov

    scipy_answer = np.sum(multivariate_normal.logpdf(
        X, np.zeros([n]), colcov_np))
    tf_answer = matnorm_logp(X_tf, rowcov, colcov)
    assert_allclose(scipy_answer, tf_answer, rtol=rtol)



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from libc.math cimport log
from numbers import Integral, Real
from typing import TypeVar, Union

import numpy as np
cimport numpy as np

np.import_array()

T = TypeVar("T", bound=Real)

def masked_log(x):
    """Compute natural logarithm while accepting nonpositive input

    For nonpositive elements, return -inf.

    Output type is the same as input type, except when input is integral, in
    which case output is `np.float64`.

    Parameters
    ----------
    x: ndarray[T]

    Returns
    -------
    ndarray[Union[T, np.float64]]
    """
    if issubclass(x.dtype.type, Integral):
        out_type = np.float64
    else:
        out_type = x.dtype
    y = np.empty(x.shape, dtype=out_type)
    lim = x.shape[0]
    for i in range(lim):
      if x[i] <= 0:
        y[i] = float('-inf')
      else:
        y[i] = log(x[i])
    return y
